"""
Compute and plot the leading EOF of geopotential height on the 500 hPa
pressure surface over the European/Atlantic sector during winter time.

This example uses the metadata-retaining xarray interface.

Additional requirements for this example:

    * xarray (http://xarray.pydata.org)
    * matplotlib (http://matplotlib.org/)
    * cartopy (http://scitools.org.uk/cartopy/)

"""
import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from eofs.xarray import Eof
from eofs.examples import example_data_path


# Read geopotential height data using the xarray module. The file contains
# December-February averages of geopotential height at 500 hPa for the
# European/Atlantic domain (80W-40E, 20-90N).
filename = example_data_path('hgt_djf.nc')
z_djf = xr.open_dataset(filename)['z']

# Compute anomalies by removing the time-mean.
z_djf = z_djf - z_djf.mean(dim='time')

# Create an EOF solver to do the EOF analysis. Square-root of cosine of
# latitude weights are applied before the computation of EOFs.
coslat = np.cos(np.deg2rad(z_djf.coords['latitude'].values)).clip(0., 1.)
wgts = np.sqrt(coslat)[..., np.newaxis]
solver = Eof(z_djf, weights=wgts)

# Retrieve the leading EOF, expressed as the covariance between the leading PC
# time series and the input SLP anomalies at each grid point.
eof1 = solver.eofsAsCovariance(neofs=1)

# Plot the leading EOF expressed as covariance in the European/Atlantic domain.
clevs = np.linspace(-75, 75, 11)
proj = ccrs.Orthographic(central_longitude=-20, central_latitude=60)
ax = plt.axes(projection=proj)
ax.coastlines()
ax.set_global()
eof1[0, 0].plot.contourf(ax=ax, levels=clevs, cmap=plt.cm.RdBu_r,
                         transform=ccrs.PlateCarree(), add_colorbar=False)
ax.set_title('EOF1 expressed as covariance', fontsize=16)
plt.show()



"""Test `eofs.multivariate` computations against reference solutions."""
# (c) Copyright 2013-2016 Andrew Dawson. All Rights Reserved.
#
# This file is part of eofs.
#
# eofs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# eofs is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with eofs.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import (absolute_import, division, print_function)  # noqa

import numpy as np
try:
    from iris.cube import Cube
except ImportError:
    pass
import pytest

import eofs.multivariate as multivariate
from eofs.tests import EofsTest

from .utils import sign_adjustments
from .reference import reference_multivariate_solution


# Create a mapping from interface name to solver class.
solvers = {'standard': multivariate.standard.MultivariateEof}
try:
    solvers['cdms'] = multivariate.cdms.MultivariateEof
except AttributeError:
    pass
try:
    solvers['iris'] = multivariate.iris.MultivariateEof
except AttributeError:
    pass


class MVSolutionTest(EofsTest):
    """Base class for all multivariate solution test classes."""
    interface = None
    weights = None
    alternate_weights_arg = None

    @classmethod
    def setup_class(cls):
        try:
            cls.solution = reference_multivariate_solution(cls.interface,
                                                           cls.weights)
        except ValueError:
            pytest.skip('missing dependencies required to test '
                        'the {!s} interface'.format(cls.interface))
        cls.neofs = cls.solution['eigenvalues'].shape[0]
        if cls.alternate_weights_arg is not None:
            weights = cls.alternate_weights_arg
        else:
            weights = cls.solution['weights']
        try:
            cls.solver = solvers[cls.interface](cls.solution['sst'],
                                                weights=weights)
        except KeyError:
            pytest.skip('missing dependencies required to test '
                        'the {!s} interface'.format(cls.interface))

    def test_eigenvalues(self):
        self.assert_array_almost_equal(
            self.solver.eigenvalues(neigs=self.neofs),
            self.solution['eigenvalues'])

    @pytest.mark.parametrize('eofscaling', (0, 1, 2))
    def test_eofs(self, eofscaling):
        eofs = [self._tomasked(e)
                for e in self.solver.eofs(neofs=self.neofs,
                                          eofscaling=eofscaling)]
        reofs = [self._tomasked(e).copy() for e in self.solution['eofs']]
        eofs = [e * sign_adjustments(e, r) for e, r in zip(eofs, reofs)]
        reigs = self._tomasked(self.solution['eigenvalues'])
        if eofscaling == 1:
            reofs = [r / np.sqrt(reigs)[:, np.newaxis, np.newaxis]
                     for r in reofs]
        elif eofscaling == 2:
            reofs = [r * np.sqrt(reigs)[:, np.newaxis, np.newaxis]
                     for r in reofs]
        for e, r in zip(eofs, reofs):
            self.assert_array_almost_equal(e, r)

    def test_eofsAsCovariance(self):
        # EOFs as covariance between PCs and input field should match the
        # reference solution
        eofs = [self._tomasked(e)
                for e in self.solver.eofsAsCovariance(neofs=self.neofs,
                                                      pcscaling=1)]
        reofs = [self._tomasked(e) for e in self.solution['eofscov']]
        eofs = [e * sign_adjustments(e, r) for e, r in zip(eofs, reofs)]
        for e, r in zip(eofs, reofs):
            self.assert_array_almost_equal(e, r)

    def test_eofsAsCorrelation(self):
        # EOFs as correlation between PCs and input field should match the
        # reference solution
        eofs = [self._tomasked(e)
                for e in self.solver.eofsAsCorrelation(neofs=self.neofs)]
        reofs = [self._tomasked(e) for e in self.solution['eofscor']]
        eofs = [e * sign_adjustments(e, r) for e, r in zip(eofs, reofs)]
        for e, r in zip(eofs, reofs):
            self.assert_array_almost_equal(e, r)

    def test_eofsAsCorrelation_range(self):
        # EOFs as correlation between PCs and input field should have values
        # in the range [-1, 1]
        eofs = [self._tomasked(e)
                for e in self.solver.eofsAsCorrelation(neofs=self.neofs)]
        for e in eofs:
            self.assert_true(np.abs(e).max() < 1.000000001)

    @pytest.mark.parametrize('pcscaling', (0, 1, 2))
    def test_pcs(self, pcscaling):
        # PCs should match the (possibly scaled) reference solution
        pcs = self._tomasked(self.solver.pcs(npcs=self.neofs,
                                             pcscaling=pcscaling))
        rpcs = self._tomasked(self.solution['pcs']).copy()
        pcs *= sign_adjustments(pcs.transpose(), rpcs.transpose()).transpose()
        reigs = self._tomasked(self.solution['eigenvalues'])
        if pcscaling == 1:
            rpcs /= np.sqrt(reigs)
        elif pcscaling == 2:
            rpcs *= np.sqrt(reigs)
        self.assert_array_almost_equal(pcs, rpcs)

    @pytest.mark.parametrize('pcscaling', (0, 1, 2))
    def test_pcs_uncorrelated(self, pcscaling):
        # PCs should be uncorrelated in time
        pcs = self._tomasked(self.solver.pcs(npcs=self.neofs,
                                             pcscaling=pcscaling))
        correlation = np.corrcoef(pcs.transpose())
        residual = correlation - np.diag(correlation.diagonal())
        self.assert_array_almost_equal(residual, 0.)

    def test_variance(self):
        # variance explained as a percentage should match the reference
        # solution
        variance = self._tomasked(
            self.solver.varianceFraction(neigs=self.neofs)) * 100.
        self.assert_array_almost_equal(variance, self.solution['variance'])

    def test_totalAnomalyVariance(self):
        # total variance should match the sum of the reference solution
        # eigenvalues
        variance = self.solver.totalAnomalyVariance()
        rvariance = self._tomasked(self.solution['eigenvalues']).sum()
        self.assert_almost_equal(variance, rvariance)

    def test_reconstructedField(self):
        sst_fields = self.solver.reconstructedField(self.neofs)
        for sst, ref in zip(sst_fields, self.solution['sst']):
            self.assert_array_almost_equal(sst, ref)

    def test_reconstructedField_arb(self):
        sst_fields = self.solver.reconstructedField([1, 2, 5])
        for sst, ref in zip(sst_fields, self.solution['rcon']):
            self.assert_array_almost_equal(sst, ref)

    def test_getWeights(self):
        # analysis weights should match those from the reference solution
        weights = self.solver.getWeights()
        if weights is not None:
            for weight, ref in zip(weights, self.solution['weights']):
                if weight is not None:
                    self.assert_array_almost_equal(weight, ref)

    @pytest.mark.parametrize('vfscaled', (True, False))
    def test_northTest(self, vfscaled):
        # typical errors should match the reference solution
        errs = self.solver.northTest(neigs=self.neofs, vfscaled=vfscaled)
        error_name = 'scaled_errors' if vfscaled else 'errors'
        self.assert_array_almost_equal(errs, self.solution[error_name])

    @pytest.mark.parametrize('eofscaling', (0, 1, 2))
    def test_projectField(self, eofscaling):
        # original input projected onto the EOFs should match the reference
        # solution PCs
        pcs = self._tomasked(self.solver.projectField(self.solution['sst'],
                                                      neofs=self.neofs,
                                                      eofscaling=eofscaling))
        rpcs = self._tomasked(self.solution['pcs']).copy()
        pcs *= sign_adjustments(pcs.transpose(), rpcs.transpose()).transpose()
        reigs = self._tomasked(self.solution['eigenvalues'])
        if eofscaling == 1:
            rpcs /= np.sqrt(reigs)
        elif eofscaling == 2:
            rpcs *= np.sqrt(reigs)
        self.assert_array_almost_equal(pcs, rpcs)

    def test_projectField_temporal_subset(self):
        # projecting a temporal subset of the original input onto the EOFs
        # should match the same subset of the reference PCs
        pcs = self._tomasked(
            self.solver.projectField([x[:5] for x in self.solution['sst']],
                                     neofs=self.neofs))
        rpcs = self._tomasked(self.solution['pcs'])[:5]
        pcs *= sign_adjustments(pcs.transpose(), rpcs.transpose()).transpose()
        self.assert_array_almost_equal(pcs, rpcs)

    def test_projectField_no_time(self):
        # projecting the first time of the original input onto the EOFs should
        # match the first time of the reference PCs
        pcs = self._tomasked(
            self.solver.projectField([x[0] for x in self.solution['sst']],
                                     neofs=self.neofs))
        rpcs = self._tomasked(self.solution['pcs'])[0]
        pcs *= sign_adjustments(pcs.transpose(), rpcs.transpose()).transpose()
        self.assert_array_almost_equal(pcs, rpcs)


# ----------------------------------------------------------------------------
# Tests for the standard interface


class StandardMVSolutionTest(MVSolutionTest):
    interface = 'standard'

    def _tomasked(self, value):
        return value


class TestStandardEqualWeights(StandardMVSolutionTest):
    weights = 'equal'


class TestStandardLatitudeWeights(StandardMVSolutionTest):
    weights = 'latitude'


class TestStandardAreaWeights(StandardMVSolutionTest):
    weights = 'area'


class TestStandardMixedWeights(StandardMVSolutionTest):
    weights = 'none_area'


# ----------------------------------------------------------------------------
# Tests for the cdms interface


class CDMSMVSolutionTest(MVSolutionTest):
    interface = 'cdms'

    def _tomasked(self, value):
        try:
            return value.asma()
        except AttributeError:
            return value


class TestCDMSEqualWeights(CDMSMVSolutionTest):
    weights = 'equal'


class TestCDMSLatitudeWeights(CDMSMVSolutionTest):
    weights = 'latitude'
    alternate_weights_arg = 'coslat'


class TestCDMSAreaWeights(CDMSMVSolutionTest):
    weights = 'area'
    alternate_weights_arg = 'area'


class TestCDMSMixedWeights(CDMSMVSolutionTest):
    weights = 'none_area'
    alternate_weights_arg = (None, 'area')


# ----------------------------------------------------------------------------
# Tests for the iris interface


class IrisMVSolutionTest(MVSolutionTest):
    interface = 'iris'

    def _tomasked(self, value):
        if type(value) is not Cube:
            return value
        return value.data


class TestIrisEqualWeights(IrisMVSolutionTest):
    weights = 'equal'


class TestIrisLatitudeWeights(IrisMVSolutionTest):
    weights = 'latitude'
    alternate_weights_arg = 'coslat'


class TestIrisAreaWeights(IrisMVSolutionTest):
    weights = 'area'
    alternate_weights_arg = 'area'


class TestIrisMixedWeights(IrisMVSolutionTest):
    weights = 'none_area'
    alternate_weights_arg = (None, 'area')



"""
Compute and plot the leading EOF of geopotential height on the 500 hPa
pressure surface over the European/Atlantic sector during winter time.

This example uses the metadata-retaining iris interface.

Additional requirements for this example:

    * iris (http://scitools.org.uk/iris/)
    * matplotlib (http://matplotlib.org/)
    * cartopy (http://scitools.org.uk/cartopy/)

"""
import warnings

import iris
import iris.plot as iplt
import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np

from eofs.iris import Eof
from eofs.examples import example_data_path


# Read geopotential height data using the iris module. The file contains
# December-February averages of geopotential height at 500 hPa for the
# European/Atlantic domain (80W-40E, 20-90N).
filename = example_data_path('hgt_djf.nc')
z_djf = iris.load_cube(filename)

# Compute anomalies by removing the time-mean.
with warnings.catch_warnings():
    # Iris emits a warning due to the non-contiguous time dimension.
    warnings.simplefilter('ignore', UserWarning)
    z_djf_mean = z_djf.collapsed('time', iris.analysis.MEAN)
z_djf = z_djf - z_djf_mean

# Create an EOF solver to do the EOF analysis. Square-root of cosine of
# latitude weights are applied before the computation of EOFs.
solver = Eof(z_djf, weights='coslat')

# Retrieve the leading EOF, expressed as the covariance between the leading PC
# time series and the input SLP anomalies at each grid point.
eof1 = solver.eofsAsCovariance(neofs=1)

# Plot the leading EOF expressed as covariance in the European/Atlantic domain.
clevs = np.linspace(-75, 75, 11)
proj = ccrs.Orthographic(central_longitude=-20, central_latitude=60)
ax = plt.axes(projection=proj)
ax.coastlines()
ax.set_global()
iplt.contourf(eof1[0, 0], levels=clevs, cmap=plt.cm.RdBu_r)
ax.set_title('EOF1 expressed as covariance', fontsize=16)
plt.show()



"""Tests for error handling in `eofs.multivariate`."""
# (c) Copyright 2013-2016 Andrew Dawson. All Rights Reserved.
#
# This file is part of eofs.
#
# eofs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# eofs is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with eofs.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import (absolute_import, division, print_function)  # noqa

import numpy as np
try:
    import cdms2
except ImportError:
    pass
import pytest

import eofs.multivariate as multivariate
from eofs.tests import EofsTest

from .reference import reference_multivariate_solution


# Create a mapping from interface name to solver class.
solvers = {'standard': multivariate.standard.MultivariateEof}
try:
    solvers['cdms'] = multivariate.cdms.MultivariateEof
except AttributeError:
    pass
try:
    solvers['iris'] = multivariate.iris.MultivariateEof
except AttributeError:
    pass


class MVErrorHandlersTest(EofsTest):
    interface = None
    weights = None

    @classmethod
    def setup_class(cls):
        try:
            cls.solution = reference_multivariate_solution(cls.interface,
                                                           cls.weights)
        except ValueError:
            pytest.skip('missing dependencies required to test '
                        'the {!s} interface'.format(cls.interface))
        cls.neofs = cls.solution['eigenvalues'].shape[0]
        try:
            cls.solver = solvers[cls.interface](
                cls.solution['sst'], weights=cls.solution['weights'])
        except KeyError:
            pytest.skip('missing dependencies required to test '
                        'the {!s} interface'.format(cls.interface))

    def test_projectField_wrong_number_fields(self):
        with pytest.raises(ValueError):
            pcs = self.solver.projectField([self.solution['sst'][0]])

    def testProjectField_time_dimension_mixture(self):
        sst1, sst2 = self.solution['sst']
        sst1 = sst1[0]
        sst2 = sst2[0:1]
        with pytest.raises(ValueError):
            pcs = self.solver.projectField([sst1, sst2])


# ----------------------------------------------------------------------------
# Error Handler tests for the standard interface


class TestErrorHandlersStandard(MVErrorHandlersTest):
    interface = 'standard'
    weights = 'equal'


# ----------------------------------------------------------------------------
# Error Handler tests for the cdms interface


class TestErrorHandlersCDMS(MVErrorHandlersTest):
    interface = 'cdms'
    weights = 'equal'

    def test_projectField_wrong_input_type(self):
        solution = reference_multivariate_solution('standard', self.weights)
        with pytest.raises(TypeError):
            pcs = self.solver.projectField(solution['sst'])

    def test_projectField_time_dimension_not_first(self):
        sst1, sst2 = self.solution['sst']
        sst1 = sst1.reorder('-t')
        with pytest.raises(ValueError):
            pcs = self.solver.projectField([sst1, sst2])


# ----------------------------------------------------------------------------
# Error Handler tests for the iris interface


class TestErrorHandlersIris(MVErrorHandlersTest):
    interface = 'iris'
    weights = 'equal'

    def test_projectField_wrong_input_type(self):
        solution = reference_multivariate_solution('standard', self.weights)
        with pytest.raises(TypeError):
            pcs = self.solver.projectField(solution['sst'])

    def test_projectField_time_dimension_not_first(self):
        sst1, sst2 = self.solution['sst']
        sst1.transpose([1, 2, 0])
        with pytest.raises(ValueError):
            pcs = self.solver.projectField([sst1, sst2])


# ----------------------------------------------------------------------------
# Constructor tests for the standard interface


class TestConstructorStandard(EofsTest):
    """Test the error handling in the standard interface constructor."""

    @classmethod
    def setup_class(cls):
        cls.solver_class = solvers['standard']

    def test_input_first_dimension_different(self):
        solution = reference_multivariate_solution('standard', 'equal')
        sst1, sst2 = solution['sst']
        sst1 = sst1[0:3]
        sst2 = sst2[0:4]
        with pytest.raises(ValueError):
            solver = self.solver_class([sst1, sst2])

    def test_wrong_number_weights(self):
        solution = reference_multivariate_solution('standard', 'area')
        weights1, weights2 = solution['weights']
        with pytest.raises(ValueError):
            solver = self.solver_class(solution['sst'], weights=[weights1])

    def test_incompatible_weights(self):
        solution = reference_multivariate_solution('standard', 'area')
        weights1, weights2 = solution['weights']
        weights2 = weights2[..., :-1]
        with pytest.raises(ValueError):
            solver = self.solver_class(solution['sst'],
                                       weights=[weights1, weights2])


# ----------------------------------------------------------------------------
# Constructor tests for the cdms interface


class TestConstructorCDMS(EofsTest):
    """Test the error handling in the cdms interface constructor."""

    @classmethod
    def setup_class(cls):
        try:
            cls.solver_class = solvers['cdms']
        except KeyError:
            pytest.skip('missing dependencies required to test '
                        'the cdms interface')

    def test_wrong_number_weights(self):
        solution = reference_multivariate_solution('cdms', 'area')
        weights1, weights2 = solution['weights']
        with pytest.raises(ValueError):
            solver = self.solver_class(solution['sst'], weights=[weights1])

    def test_wrong_input_type(self):
        solution = reference_multivariate_solution('standard', 'equal')
        with pytest.raises(TypeError):
            solver = self.solver_class(solution['sst'])

    def test_input_time_dimension_not_first(self):
        solution = reference_multivariate_solution('cdms', 'equal')
        sst1, sst2 = solution['sst']
        sst1 = sst1.reorder('-t')
        with pytest.raises(ValueError):
            solver = self.solver_class([sst1, sst2])

    def test_input_no_spatial_dimensions(self):
        solution = reference_multivariate_solution('cdms', 'equal')
        sst1, sst2 = solution['sst']
        sst1 = sst1[:, 0, 0]
        with pytest.raises(ValueError):
            solver = self.solver_class([sst1, sst2])


# ----------------------------------------------------------------------------
# Constructor tests for the standard interface


class TestConstructorIris(EofsTest):
    """Test the error handling in the iris interface constructor."""

    @classmethod
    def setup_class(cls):
        try:
            cls.solver_class = solvers['iris']
        except KeyError:
            pytest.skip('missing dependencies required to test '
                        'the iris interface')

    def test_wrong_number_weights(self):
        solution = reference_multivariate_solution('iris', 'area')
        weights1, weights2 = solution['weights']
        with pytest.raises(ValueError):
            solver = self.solver_class(solution['sst'], weights=[weights1])

    def test_wrong_input_type(self):
        solution = reference_multivariate_solution('standard', 'equal')
        with pytest.raises(TypeError):
            solver = self.solver_class(solution['sst'])

    def test_input_time_dimension_not_first(self):
        solution = reference_multivariate_solution('iris', 'equal')
        sst1, sst2 = solution['sst']
        sst1.transpose([1, 2, 0])
        with pytest.raises(ValueError):
            solver = self.solver_class([sst1, sst2])

    def test_input_no_spatial_dimensions(self):
        solution = reference_multivariate_solution('iris', 'equal')
        sst1, sst2 = solution['sst']
        sst1 = sst1[:, 0, 0]
        with pytest.raises(ValueError):
            solver = self.solver_class([sst1, sst2])



"""Reference solutions for testing the `eofs` package."""
# (c) Copyright 2013-2016 Andrew Dawson. All Rights Reserved.
#
# This file is part of eofs.
#
# eofs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# eofs is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with eofs.  If not, see <http://www.gnu.org/licenses/>.
import os

import numpy as np
import numpy.ma as ma


def _test_data_path():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), 'data'))


def _retrieve_test_field(name):
    filename = os.path.join(_test_data_path(), '{!s}.npy'.format(name))
    try:
        field = np.load(filename)
    except IOError:
        field = None
    return field


def _tomasked(array):
    try:
        return ma.MaskedArray(array, mask=np.isnan(array))
    except TypeError:
        return array


def _read_reference_solution(weights):
    if weights not in ('equal', 'latitude', 'area', 'area_multi',
                       'area_multi_mix'):
        raise ValueError("invalid weights: '{!s}'".format(weights))
    field_names = ['time',
                   'latitude',
                   'longitude',
                   'sst',
                   'eigenvalues.{!s}'.format(weights),
                   'eofs.{!s}'.format(weights),
                   'eofscor.{!s}'.format(weights),
                   'eofscov.{!s}'.format(weights),
                   'pcs.{!s}'.format(weights),
                   'variance.{!s}'.format(weights),
                   'weights.{!s}'.format(weights),
                   'errors.{!s}'.format(weights),
                   'scaled_errors.{!s}'.format(weights),
                   'rcon.{!s}'.format(weights)]
    fields = {name.split('.')[0]: _tomasked(_retrieve_test_field(name))
              for name in field_names}
    fields['sst'] -= fields['sst'].mean(axis=0)
    return fields


def _wrap_cdms(solution, neofs, time_units):
    try:
        import cdms2
    except ImportError:
        raise ValueError("cannot use container 'cdms' without "
                         "the cdms2 module")
    time_dim = cdms2.createAxis(solution['time'], id='time')
    time_dim.designateTime()
    time_dim.units = time_units
    lat_dim = cdms2.createAxis(solution['latitude'], id='latitude')
    lat_dim.designateLatitude()
    lon_dim = cdms2.createAxis(solution['longitude'], id='longitude')
    lon_dim.designateLongitude()
    eof_dim = cdms2.createAxis(np.arange(1, neofs+1), id='eof')
    eof_dim.long_name = 'eof_number'
    solution['sst'] = cdms2.createVariable(
        solution['sst'],
        axes=[time_dim, lat_dim, lon_dim],
        id='sst')
    solution['eigenvalues'] = cdms2.createVariable(
        solution['eigenvalues'],
        axes=[eof_dim],
        id='eigenvalues')
    solution['eofs'] = cdms2.createVariable(
        solution['eofs'],
        axes=[eof_dim, lat_dim, lon_dim],
        id='eofs')
    solution['pcs'] = cdms2.createVariable(
        solution['pcs'],
        axes=[time_dim, eof_dim],
        id='pcs')
    solution['variance'] = cdms2.createVariable(
        solution['variance'],
        axes=[eof_dim],
        id='variance')
    solution['eofscor'] = cdms2.createVariable(
        solution['eofscor'],
        axes=[eof_dim, lat_dim, lon_dim],
        id='eofscor')
    solution['eofscov'] = cdms2.createVariable(
        solution['eofscov'],
        axes=[eof_dim, lat_dim, lon_dim],
        id='eofscov')
    solution['errors'] = cdms2.createVariable(
        solution['errors'],
        axes=[eof_dim],
        id='errors')
    solution['scaled_errors'] = cdms2.createVariable(
        solution['scaled_errors'],
        axes=[eof_dim],
        id='scaled_errors')
    solution['rcon'] = cdms2.createVariable(
        solution['rcon'],
        axes=[time_dim, lat_dim, lon_dim],
        id='reconstructed_sst')


def _wrap_iris(solution, neofs, time_units):
    try:
        from iris.cube import Cube
        from iris.coords import DimCoord
        from cf_units import Unit
    except ImportError:
        raise ValueError("cannot use container 'iris' without "
                         "the iris module")
    time_dim = DimCoord(solution['time'],
                        standard_name='time',
                        units=Unit(time_units, 'gregorian'))
    lat_dim = DimCoord(solution['latitude'],
                       standard_name='latitude',
                       units='degrees_north')
    lat_dim.guess_bounds()
    lon_dim = DimCoord(solution['longitude'],
                       standard_name='longitude',
                       units='degrees_east')
    lon_dim.guess_bounds()
    eof_dim = DimCoord(np.arange(1, neofs+1),
                       long_name='eof')
    solution['sst'] = Cube(
        solution['sst'],
        dim_coords_and_dims=list(zip((time_dim, lat_dim, lon_dim), range(3))),
        long_name='sst')
    solution['eigenvalues'] = Cube(
        solution['eigenvalues'],
        dim_coords_and_dims=list(zip((eof_dim,), range(1))),
        long_name='eigenvalues')
    solution['eofs'] = Cube(
        solution['eofs'],
        dim_coords_and_dims=list(zip((eof_dim, lat_dim, lon_dim), range(3))),
        long_name='eofs')
    solution['pcs'] = Cube(
        solution['pcs'],
        dim_coords_and_dims=list(zip((time_dim, eof_dim), range(2))),
        long_name='pcs')
    solution['variance'] = Cube(
        solution['variance'],
        dim_coords_and_dims=list(zip((eof_dim,), range(1))),
        long_name='variance')
    solution['eofscor'] = Cube(
        solution['eofscor'],
        dim_coords_and_dims=list(zip((eof_dim, lat_dim, lon_dim), range(3))),
        long_name='eofscor')
    solution['eofscov'] = Cube(
        solution['eofscov'],
        dim_coords_and_dims=list(zip((eof_dim, lat_dim, lon_dim), range(3))),
        long_name='eofscov')
    solution['errors'] = Cube(
        solution['errors'],
        dim_coords_and_dims=list(zip((eof_dim,), range(1))),
        long_name='errors')
    solution['scaled_errors'] = Cube(
        solution['scaled_errors'],
        dim_coords_and_dims=list(zip((eof_dim,), range(1))),
        long_name='scaled_errors')
    solution['rcon'] = Cube(
        solution['rcon'],
        dim_coords_and_dims=list(zip((time_dim, lat_dim, lon_dim), range(3))),
        long_name='reconstructed_sst')


def _wrap_xarray(solution, neofs, time_units):
    try:
        import xarray as xr
    except ImportError:
        try:
            import xray as xr
        except ImportError:
            raise ValueError("cannot use container 'xarray' without "
                             "the xarray/xray module")
    time_dim = xr.IndexVariable('time', solution['time'])
    lat_dim = xr.IndexVariable('latitude', solution['latitude'])
    lon_dim = xr.IndexVariable('longitude', solution['longitude'])
    eof_dim = xr.IndexVariable('eof', np.arange(1, neofs+1))
    solution['sst'] = xr.DataArray(solution['sst'],
                                   coords=[time_dim, lat_dim, lon_dim])
    solution['eigenvalues'] = xr.DataArray(solution['eigenvalues'],
                                           coords=[eof_dim])
    solution['eofs'] = xr.DataArray(solution['eofs'],
                                    coords=[eof_dim, lat_dim, lon_dim])


def _get_wrapper(container_type):
    return {'standard': lambda *args: None,
            'cdms': _wrap_cdms,
            'iris': _wrap_iris,
            'xarray': _wrap_xarray}[container_type]


def reference_solution(container_type, weights):
    """Obtain a reference EOF analysis solution.

    **Arguments:**

    *container_type*
        Either 'standard', 'cdms', 'iris' or 'xarray'.

    *weights*
        Weights method. One of 'equal', 'latitude', or 'area'.

    """
    container_type = container_type.lower()
    weights = weights.lower()
    if container_type not in ('standard', 'iris', 'cdms', 'xarray'):
        raise ValueError("unknown container type "
                         "'{!s}'".format(container_type))
    solution = _read_reference_solution(weights)
    time_units = 'months since 0-1-1 00:00:0.0'
    neofs = len(solution['eigenvalues'])
    _get_wrapper(container_type)(solution, neofs, time_units)
    return solution


def reference_multivariate_solution(container_type, weights):
    """Obtain a reference multivariate EOF analysis solution.

    **Arguments:**

    *container_type*
        Either 'standard', 'cdms', or 'iris'.

    *weights*
        Weights method. One of 'equal', 'latitude', 'area',
        'area_multi', or 'area_multi_mix'.

    """
    if weights.lower() == 'area':
        weights = 'area_multi'
    if weights.lower() == 'none_area':
        weights = 'area_multi_mix'
    solution = reference_solution(container_type, weights)
    nlon = len(solution['longitude'])
    slice1 = slice(0, nlon // 2)
    slice2 = slice(nlon // 2, None)
    for var in ('longitude',
                'sst',
                'eofs',
                'eofscor',
                'eofscov',
                'weights',
                'rcon',):
        try:
            solution[var] = (solution[var][..., slice1],
                             solution[var][..., slice2])
        except TypeError:
            solution[var] = None, None
    return solution



"""Supplementary tools for the `xarray` EOF analysis interface."""
# (c) Copyright 2016 Andrew Dawson. All Rights Reserved.
#
# This file is part of eofs.
#
# eofs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# eofs is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with eofs.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import (absolute_import, division, print_function)  # noqa

import numpy as np
try:
    import xarray as xr
except ImportError:
    import xray as xr

from . import standard
from .generic import covcor_dimensions


def find_time_coordinates(array):
    """Find time dimension coordinates in an `xarray.DataArray`.

    Time coordinates satisfy one or more of:

    * Values have a dtype of `numpy.datetime64`.
    * Name of the coordinate is 'time'.
    * The coordinate has an attribute 'axis' with value 'T'.

    **Argument:**

    *array*
        An `xarray.DataArray`.

    **Returns:**

    *time_coords*
        A list of time dimension coordinates. The list may be empty if
        no coordinates were recognised as time.

    """
    time_coords = []
    for name in array.dims:
        coord = array.coords[name]
        is_time = (np.issubdtype(coord.dtype, np.datetime64) or
                   coord.name == 'time' or
                   coord.attrs.get('axis') == 'T')
        if is_time:
            time_coords.append(coord)
    return time_coords


def categorise_ndcoords(array, time_name):
    """
    Categorise all the non-dimension coordinates of an
    `xarray.DataArray` into those that span only time, those that span
    only space, and those that span both time and space.

    **Arguments:**

    *array*
        An `xarray.DataArray`.

    *time_name*
        Name of the time dimension coordinate in the input *array*.

    **Returns:**

    *time_coords*
        A list of coordinates that span only the time dimension.

    *space_coords*
        A list of coordinates that span only the space dimensions.

    *time_space_coords*
        A list of coordinates that span both the time and space
        coordinates.

    """
    ndcoords = [coord for name, coord in array.coords.items()
                if name not in array.dims]
    time_ndcoords = []
    space_ndcoords = []
    time_space_ndcoords = []
    for coord in ndcoords:
        if coord.dims == (time_name,):
            time_ndcoords.append(coord)
        elif coord.dims:
            if time_name in coord.dims:
                time_space_ndcoords.append(coord)
            else:
                space_ndcoords.append(coord)
    return time_ndcoords, space_ndcoords, time_space_ndcoords


def weights_array(array, scheme):
    """Generate a weights array for a given weighting scheme.

    .. note::

       Currently not implemented, will raise a ValueError().

    """
    scheme = scheme.lower()
    raise ValueError("invalid weighting scheme: '{!s}'".format(scheme))


def _coord_info(array):
    time_coords = find_time_coordinates(array)
    if len(time_coords) > 1:
        raise ValueError('multiple time coordinates are not allowed')
    if not time_coords:
        msg = 'no time coordinates found in {!s}'.format(array.name)
        raise ValueError(msg)
    time_coord = time_coords[0]
    time_dim = array.dims.index(time_coord.name)
    coords = [time_coord] + [array.coords[name] for name in array.dims
                             if name != time_coord.name]
    return time_dim, coords


def _map_and_coords(pcs, field, mapfunc, *args, **kwargs):
    info = {}
    for array in (field, pcs):
        info[array.name] = _coord_info(array)
    cmap_args = [np.rollaxis(array.values, info[array.name][0])
                 for array in (pcs, field)]
    cmap_args += args
    cmap = mapfunc(*cmap_args, **kwargs)
    dim_args = [info[array.name][1] for array in (pcs, field)]
    dims = covcor_dimensions(*dim_args)
    return cmap, dims


def correlation_map(pcs, field):
    """Correlation between PCs and a field.

    Computes maps of the correlation between each PC and the given
    field at each grid point.

    Given a set of PCs contained in a `~xarray.DataArray` (e.g., as
    output from `eofs.xarray.Eof.pcs`) and a field with a time dimension
    contained in a `xarray.DataArray`, one correlation map per PC is
    computed.

    The field must have the same length time dimension as the PCs. Any
    number of spatial dimensions (including zero) are allowed in the
    field and there can be any number of PCs.

    **Arguments:**

    *pcs*
        PCs contained in a `~xarray.DataArray`.

    *field*
        Field Spatial-temporal field contained in a `~xarray.DataArray`.

    **Returns:**

    *correlation_maps*
        A `~xarray.DataArray` containing the correlation maps.

    **Examples:**

    Assuming *solver* is an instance of `eofs.xarray.Eof`, compute
    correlation maps for each PC::

        pcs = solver.pcs(pcscaling=1)
        correlation_maps = correlation_map(pcs, field)

    """
    cor, coords = _map_and_coords(pcs, field, standard.correlation_map)
    if not coords:
        return cor
    cor = xr.DataArray(cor, coords=coords, name='pc_correlation',
                       attrs={'long_name': 'pc_correlation'})
    return cor


def covariance_map(pcs, field, ddof=1):
    """Covariance between PCs and a field.

    Computes maps of the covariance between each PC and the given
    field at each grid point.

    Given a set of PCs contained in a `~xarray.DataArray` (e.g., as
    output from `eofs.xarray.Eof.pcs`) and a field with a time dimension
    contained in a `xarray.DataArray`, one covariance map per PC is
    computed.

    The field must have the same length time dimension as the PCs. Any
    number of spatial dimensions (including zero) are allowed in the
    field and there can be any number of PCs.

    **Arguments:**

    *pcs*
        PCs contained in a `~xarray.DataArray`.

    *field*
        Field Spatial-temporal field contained in a `~xarray.DataArray`.

    **Keyword arguments:**

    *ddof*
        'Delta degrees of freedom'. The divisor used to normalize
        the covariance matrix is *N - ddof* where *N* is the
        number of samples. Defaults to *1*.

    **Returns:**

    *covariance_maps*
        A `~xarray.DataArray` containing the covariance maps.

    **Examples:**

    Assuming *solver* is an instance of `eofs.xarray.Eof`, compute
    covariance maps for each PC::

        pcs = solver.pcs(pcscaling=1)
        covariance_maps = covariance_map(pcs, field)

    """
    cov, coords = _map_and_coords(pcs, field, standard.covariance_map,
                                  ddof=ddof)
    if not coords:
        return cov
    cov = xr.DataArray(cov, coords=coords, name='pc_covariance',
                       attrs={'long_name': 'pc_covariance'})
    return cov



"""
Compute and plot the leading EOF of geopotential height on the 500 hPa
pressure surface over the European/Atlantic sector during winter time.

This example uses the metadata-retaining cdms2 interface.


Additional requirements for this example:

    * cdms2 (http://uvcdat.llnl.gov/)
    * matplotlib (http://matplotlib.org/)
    * cartopy (http://scitools.org.uk/cartopy/)

"""
import cartopy.crs as ccrs
import cdms2
import cdutil
import matplotlib.pyplot as plt
import numpy as np

from eofs.cdms import Eof
from eofs.examples import example_data_path


# Read geopotential height data using the cdms2 module from UV-CDAT. The file
# contains December-February averages of geopotential height at 500 hPa for
# the European/Atlantic domain (80W-40E, 20-90N).
filename = example_data_path('hgt_djf.nc')
ncin = cdms2.open(filename, 'r')
z_djf = ncin('z')
ncin.close()

# Compute anomalies by removing the time-mean.
z_djf_mean = cdutil.averager(z_djf, axis='t')
z_djf = z_djf - z_djf_mean
z_djf.id = 'z'

# Create an EOF solver to do the EOF analysis. Square-root of cosine of
# latitude weights are applied before the computation of EOFs.
solver = Eof(z_djf, weights='coslat')

# Retrieve the leading EOF, expressed as the covariance between the leading PC
# time series and the input SLP anomalies at each grid point.
eof1 = solver.eofsAsCovariance(neofs=1)

# Plot the leading EOF expressed as covariance in the European/Atlantic domain.
clevs = np.linspace(-75, 75, 11)
lons, lats = eof1.getLongitude()[:], eof1.getLatitude()[:]
proj = ccrs.Orthographic(central_longitude=-20, central_latitude=60)
ax = plt.axes(projection=proj)
ax.set_global()
ax.coastlines()
ax.contourf(lons, lats, eof1(squeeze=True).data, levels=clevs,
            cmap=plt.cm.RdBu_r, transform=ccrs.PlateCarree())
plt.title('EOF1 expressed as covariance', fontsize=16)

plt.show()



"""Meta-data preserving EOF analysis for `iris`."""
# (c) Copyright 2013-2016 Andrew Dawson. All Rights Reserved.
#
# This file is part of eofs.
#
# eofs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# eofs is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with eofs.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import (absolute_import, division, print_function)  # noqa

import collections.abc
from copy import copy

from iris.cube import Cube
from iris.coords import DimCoord

from . import standard
from .tools.iris import get_time_coord, weights_array, classified_aux_coords


class Eof(object):
    """EOF analysis (meta-data enabled `iris` interface)"""

    def __init__(self, cube, weights=None, center=True, ddof=1):
        """Create an Eof object.

        The EOF solution is computed at initialization time. Method
        calls are used to retrieve computed quantities.

        **Argument:**

        *dataset*
            A `~iris.cube.Cube` instance containing the data to be
            analysed. Time must be the first dimension. Missing values
            are allowed provided that they are constant with time (e.g.,
            values of an oceanographic field over land).

        **Optional arguments:**

        *weights*
            Sets the weighting method. The following pre-defined
            weighting methods are available:

            * *'area'* : Square-root of grid cell area normalized by
              total grid area. Requires a latitude-longitude grid to be
              present in the `~iris.cube.Cube` *dataset*. This is a
              fairly standard weighting strategy. If you are unsure
              which method to use and you have gridded data then this
              should be your first choice.

            * *'coslat'* : Square-root of cosine of latitude. Requires a
              latitude dimension to be present in the `~iris.cube.Cube`
              *dataset*.

            * *None* : Equal weights for all grid points (*'none'* is
              also accepted).

             Alternatively an array of weights whose shape is compatible
             with the `~iris.cube.Cube` *dataset* may be supplied instead
             of specifying a weighting method.

        *center*
            If *True*, the mean along the first axis of *dataset* (the
            time-mean) will be removed prior to analysis. If *False*,
            the mean along the first axis will not be removed. Defaults
            to *True* (mean is removed).

            The covariance interpretation relies on the input data being
            anomalies with a time-mean of 0. Therefore this option
            should usually be set to *True*. Setting this option to
            *True* has the useful side effect of propagating missing
            values along the time dimension, ensuring that a solution
            can be found even if missing values occur in different
            locations at different times.

        *ddof*
            'Delta degrees of freedom'. The divisor used to normalize
            the covariance matrix is *N - ddof* where *N* is the
            number of samples. Defaults to *1*.

        **Returns:**

        *solver*
            An `Eof` instance.

        **Examples:**

        EOF analysis with grid-cell-area weighting for the input field::

            from eofs.iris import Eof
            solver = Eof(cube, weights='area')

        """
        # Check that the input is an Iris cube.
        if not isinstance(cube, Cube):
            raise TypeError('the input must be an iris cube')
        # Check for a time coordinate, raise an error if there isn't one.
        # The get_time_coord function will raise a ValuerError with a
        # useful message so no need to handle it explicitly here.
        _time, self._time_dim = get_time_coord(cube)
        self._time = copy(_time)
        if self._time_dim != 0:
            raise ValueError('time must be the first dimension, '
                             'consider using the transpose() method')
        # Get the cube coordinates and remove time, leaving just the other
        # dimensions.
        self._coords = [copy(coord) for coord in cube.dim_coords]
        self._coords.remove(self._time)
        if not self._coords:
            raise ValueError('one or more non-time dimensions are required')
        # Store the auxiliary coordinates from the cube, categorising them into
        # coordinates spanning time only, coordinates spanning space only, and
        # coordinates spanning both time and space. This is helpful due to the
        # natural separation of space and time in EOF analysis. The time and
        # space spanning coordinates are only useful for reconstruction, as all
        # other methods return either a temporal field or a spatial field.
        (self._time_aux_coords,
         self._space_aux_coords,
         self._time_space_aux_coords) = classified_aux_coords(cube)
        # Define the weights array for the cube.
        if weights is None:
            wtarray = None
        else:
            try:
                scheme = weights.lower()
                wtarray = weights_array(cube, scheme=scheme)
            except AttributeError:
                wtarray = weights
        try:
            # Ensure weights are the same type as the cube data.
            wtarray = wtarray.astype(cube.data.dtype)
        except AttributeError:
            pass
        # Initialize a solver.
        self._solver = standard.Eof(cube.data,
                                    weights=wtarray,
                                    center=center,
                                    ddof=ddof)
        #: Number of EOFs in the solution.
        self.neofs = self._solver.neofs
        # Get the name of the cube to refer to later.
        self._cube_name = cube.name(default='dataset').replace(' ', '_')
        self._cube_var_name = cube.var_name

    def pcs(self, pcscaling=0, npcs=None):
        """Principal component time series (PCs).

        **Optional arguments:**

        *pcscaling*
            Set the scaling of the retrieved PCs. The following
            values are accepted:

            * *0* : Un-scaled principal components (default).
            * *1* : Principal components are scaled to unit variance
              (divided by the square-root of their eigenvalue).
            * *2* : Principal components are multiplied by the
              square-root of their eigenvalue.

        *npcs*
            Number of PCs to retrieve. Defaults to all the PCs. If the
            number of requested PCs is more than the number that are
            available, then all available PCs will be returned.

        **Returns:**

        *pcs*
            A `~iris.cube.Cube` containing the ordered PCs. The PCs are
            numbered from 0 to *npcs* - 1.

        **Examples:**

        All un-scaled PCs::

            pcs = solver.pcs()

        First 3 PCs scaled to unit variance::

            pcs = solver.pcs(npcs=3, pcscaling=1)

        """
        pcs = self._solver.pcs(pcscaling, npcs)
        pcdim = DimCoord(list(range(pcs.shape[1])),
                         var_name='pc',
                         long_name='pc_number')
        coords = [copy(self._time), pcdim]
        pcs = Cube(
            pcs,
            dim_coords_and_dims=list(zip(coords, list(range(pcs.ndim)))),
            var_name='pcs',
            long_name='principal_components')
        # Add any auxiliary coords spanning time back to the returned cube.
        for coord, dims in self._time_aux_coords:
            pcs.add_aux_coord(copy(coord), dims)
        return pcs

    def eofs(self, eofscaling=0, neofs=None):
        """Emipirical orthogonal functions (EOFs).

        **Optional arguments:**

        *eofscaling*
            Sets the scaling of the EOFs. The following values are
            accepted:

            * *0* : Un-scaled EOFs (default).
            * *1* : EOFs are divided by the square-root of their
              eigenvalues.
            * *2* : EOFs are multiplied by the square-root of their
              eigenvalues.

        *neofs*
            Number of EOFs to return. Defaults to all EOFs. If the
            number of EOFs requested is more than the number that are
            available, then all available EOFs will be returned.

        **Returns:**

        *eofs*
           A `~iris.cube.Cube` containing the ordered EOFs. The EOFs are
           numbered from 0 to *neofs* - 1.

        **Examples:**

        All EOFs with no scaling::

            eofs = solver.eofs()

        First 3 EOFs with scaling applied::

            eofs = solver.eofs(neofs=3, eofscaling=1)

        """
        eofs = self._solver.eofs(eofscaling, neofs)
        eofdim = DimCoord(list(range(eofs.shape[0])),
                          var_name='eof',
                          long_name='eof_number')
        coords = [eofdim] + [copy(coord) for coord in self._coords]
        eofs = Cube(
            eofs,
            dim_coords_and_dims=list(zip(coords, list(range(eofs.ndim)))),
            var_name='eofs',
            long_name='empirical_orthogonal_functions')
        # Add any auxiliary coordinates spanning space to the returned cube.
        for coord, dims in self._space_aux_coords:
            eofs.add_aux_coord(copy(coord), dims)
        return eofs

    def eofsAsCorrelation(self, neofs=None):
        """
        Empirical orthogonal functions (EOFs) expressed as the
        correlation between the principal component time series (PCs)
        and the time series of the `Eof` input *dataset* at each grid
        point.

        .. note::

            These are not related to the EOFs computed from the
            correlation matrix.

        **Optional argument:**

        *neofs*
            Number of EOFs to return. Defaults to all EOFs. If the
            number of EOFs requested is more than the number that are
            available, then all available EOFs will be returned.

        **Returns:**

        *eofs*
           A `~iris.cube.Cube` containing the ordered EOFs. The EOFs are
           numbered from 0 to *neofs* - 1.

        **Examples:**

        All EOFs::

            eofs = solver.eofsAsCorrelation()

        The leading EOF::

            eof1 = solver.eofsAsCorrelation(neofs=1)

        """
        eofs = self._solver.eofsAsCorrelation(neofs)
        eofdim = DimCoord(list(range(eofs.shape[0])),
                          var_name='eof',
                          long_name='eof_number')
        coords = [eofdim] + [copy(coord) for coord in self._coords]
        eofs = Cube(
            eofs,
            dim_coords_and_dims=list(zip(coords, list(range(eofs.ndim)))),
            var_name='eofs',
            long_name='correlation_between_pcs_and_{:s}'.format(
                self._cube_name))
        # Add any auxiliary coordinates spanning space to the returned cube.
        for coord, dims in self._space_aux_coords:
            eofs.add_aux_coord(copy(coord), dims)
        return eofs

    def eofsAsCovariance(self, neofs=None, pcscaling=1):
        """
        Empirical orthogonal functions (EOFs) expressed as the
        covariance between the principal component time series (PCs)
        and the time series of the `Eof` input *dataset* at each grid
        point.

        **Optional arguments:**

        *neofs*
            Number of EOFs to return. Defaults to all EOFs. If the
            number of EOFs requested is more than the number that are
            available, then all available EOFs will be returned.

        *pcscaling*
            Set the scaling of the PCs used to compute covariance. The
            following values are accepted:

            * *0* : Un-scaled PCs.
            * *1* : PCs are scaled to unit variance (divided by the
              square-root of their eigenvalue) (default).
            * *2* : PCs are multiplied by the square-root of their
              eigenvalue.

            The default is to divide PCs by the square-root of their
            eigenvalue so that the PCs are scaled to unit variance
            (option 1).

        **Returns:**

        *eofs*
           A `~iris.cube.Cube` containing the ordered EOFs. The EOFs are
           numbered from 0 to *neofs* - 1.

        **Examples:**

        All EOFs::

            eofs = solver.eofsAsCovariance()

        The leading EOF::

            eof1 = solver.eofsAsCovariance(neofs=1)

        The leading EOF using un-scaled PCs::

            eof1 = solver.eofsAsCovariance(neofs=1, pcscaling=0)

        """
        eofs = self._solver.eofsAsCovariance(neofs, pcscaling)
        eofdim = DimCoord(list(range(eofs.shape[0])),
                          var_name='eof',
                          long_name='eof_number')
        coords = [eofdim] + [copy(coord) for coord in self._coords]
        eofs = Cube(
            eofs,
            dim_coords_and_dims=list(zip(coords, list(range(eofs.ndim)))),
            var_name='eofs',
            long_name='covariance_between_pcs_and_{:s}'.format(
                self._cube_name))
        # Add any auxiliary coordinates spanning space to the returned cube.
        for coord, dims in self._space_aux_coords:
            eofs.add_aux_coord(copy(coord), dims)
        return eofs

    def eigenvalues(self, neigs=None):
        """Eigenvalues (decreasing variances) associated with each EOF.

        **Optional argument:**

        *neigs*
            Number of eigenvalues to return. Defaults to all
            eigenvalues.If the number of eigenvalues requested is more
            than the number that are available, then all available
            eigenvalues will be returned.

        **Returns:**

        *eigenvalues*
            A `~iris.cube.Cube` containing the eigenvalues arranged
            largest to smallest. The eigenvalues are numbered from 0 to
            *neigs* - 1.

        **Examples:**

        All eigenvalues::

            eigenvalues = solver.eigenvalues()

        The first eigenvalue::

            eigenvalue1 = solver.eigenvalues(neigs=1)

        """
        lambdas = self._solver.eigenvalues(neigs=neigs)
        eofdim = DimCoord(list(range(lambdas.shape[0])),
                          var_name='eigenvalue',
                          long_name='eigenvalue_number')
        coords = [eofdim]
        lambdas = Cube(
            lambdas,
            dim_coords_and_dims=list(zip(coords, list(range(lambdas.ndim)))),
            var_name='eigenvalues',
            long_name='eigenvalues')
        return lambdas

    def varianceFraction(self, neigs=None):
        """Fractional EOF mode variances.

        The fraction of the total variance explained by each EOF mode,
        values between 0 and 1 inclusive.

        **Optional argument:**

        *neigs*
            Number of eigenvalues to return the fractional variance for.
            Defaults to all eigenvalues. If the number of eigenvalues
            requested is more than the number that are available, then
            fractional variances for all available eigenvalues will be
            returned.

        **Returns:**

        *variance_fractions*
            A `~iris.cube.Cube` containing the fractional variances for
            each eigenvalue. The eigenvalues are numbered from 0 to
            *neigs* - 1.


        **Examples:**

        The fractional variance represented by each eigenvalue::

            variance_fractions = solver.varianceFraction()

        The fractional variance represented by the first 3 eigenvalues::

            variance_fractions = solver.VarianceFraction(neigs=3)

        """
        vfrac = self._solver.varianceFraction(neigs=neigs)
        eofdim = DimCoord(list(range(vfrac.shape[0])),
                          var_name='eigenvalue',
                          long_name='eigenvalue_number')
        coords = [eofdim]
        vfrac = Cube(
            vfrac,
            dim_coords_and_dims=list(zip(coords, list(range(vfrac.ndim)))),
            var_name='variance_fractions',
            long_name='variance_fractions')
        return vfrac

    def totalAnomalyVariance(self):
        """
        Total variance associated with the field of anomalies (the sum
        of the eigenvalues).

        **Returns:**

        *total_variance*
            A scalar value (not a `~iris.cube.Cube`).

        **Example:**

        Get the total variance::

            total_variance = solver.totalAnomalyVariance()

        """
        return self._solver.totalAnomalyVariance()

    def northTest(self, neigs=None, vfscaled=False):
        """Typical errors for eigenvalues.

        The method of North et al. (1982) is used to compute the typical
        error for each eigenvalue. It is assumed that the number of
        times in the input data set is the same as the number of
        independent realizations. If this assumption is not valid then
        the result may be inappropriate.

        **Optional arguments:**

        *neigs*
            The number of eigenvalues to return typical errors for.
            Defaults to typical errors for all eigenvalues.

        *vfscaled*
            If *True* scale the errors by the sum of the eigenvalues.
            This yields typical errors with the same scale as the values
            returned by `Eof.varianceFraction`. If *False* then no
            scaling is done. Defaults to *False* (no scaling).

        **Returns:**

        *errors*
            A `~iris.cube.Cube` containing the typical errors for each
            eigenvalue. The egienvalues are numbered from 0 to
            *neigs* - 1.

        **References**

        North G.R., T.L. Bell, R.F. Cahalan, and F.J. Moeng (1982)
        Sampling errors in the estimation of empirical orthogonal
        functions. *Mon. Weather. Rev.*, **110**, pp 669-706.

        **Examples:**

        Typical errors for all eigenvalues::

            errors = solver.northTest()

        Typical errors for the first 3 eigenvalues scaled by the sum of
        the eigenvalues::

            errors = solver.northTest(neigs=3, vfscaled=True)

        """
        typerrs = self._solver.northTest(neigs=neigs, vfscaled=vfscaled)
        eofdim = DimCoord(list(range(typerrs.shape[0])),
                          var_name='eigenvalue',
                          long_name='eigenvalue_number')
        coords = [eofdim]
        typerrs = Cube(
            typerrs,
            dim_coords_and_dims=list(zip(coords, list(range(typerrs.ndim)))),
            var_name='typical_errors',
            long_name='typical_errors')
        return typerrs

    def reconstructedField(self, neofs):
        """Reconstructed data field based on a subset of EOFs.

        If weights were passed to the `Eof` instance the returned
        reconstructed field will automatically have this weighting
        removed. Otherwise the returned field will have the same
        weighting as the `Eof` input *dataset*.

        Returns the reconstructed field in a `~iris.cube.Cube`.

        **Argument:**

        *neofs*
            Number of EOFs to use for the reconstruction.
            Alternatively this argument can be an iterable of mode
            numbers (where the first mode is 1) in order to facilitate
            reconstruction with arbitrary modes.

        **Returns:**

        *reconstruction*
            A `~iris.cube.Cube` with the same dimensions `Eof` input
            *dataset* containing the reconstruction using *neofs* EOFs.

        **Example:**

        Reconstruct the input field using 3 EOFs::

            reconstruction = solver.reconstructedField(3)

        Reconstruct the input field using EOFs 1, 2 and 5::

            reconstruction = solver.reconstuctedField([1, 2, 5])

        """
        rfield = self._solver.reconstructedField(neofs)
        coords = [copy(self._time)] + [copy(coord) for coord in self._coords]
        if isinstance(neofs, collections.abc.Iterable):
            name_part = 'EOFs_{}'.format('_'.join([str(e) for e in neofs]))
        else:
            name_part = '{}_EOFs'.format(neofs)
        rfield = Cube(
            rfield,
            dim_coords_and_dims=list(zip(coords, list(range(rfield.ndim)))),
            var_name=self._cube_var_name or 'dataset',
            long_name='{:s}_reconstructed_with_{:s}'.format(
                self._cube_name, name_part))
        rfield.attributes.update({'neofs': neofs})
        # Add any auxiliary coordinates to the returned cube.
        for coord, dims in (self._time_aux_coords +
                            self._space_aux_coords +
                            self._time_space_aux_coords):
            rfield.add_aux_coord(copy(coord), dims)
        return rfield

    def projectField(self, cube, neofs=None, eofscaling=0, weighted=True):
        """Project a field onto the EOFs.

        Given a data set, projects it onto the EOFs to generate a
        corresponding set of pseudo-PCs.

        **Argument:**

        *field*
            An `iris.cube.Cube` containing the field to project onto the
            EOFs. It must have the same corresponding spatial dimensions
            (including missing values in the same places) as the `Eof`
            input *dataset*. It may have a different length time
            dimension to the `Eof` input *dataset* or no time dimension
            at all. If a time dimension exists it must be the first
            dimension.

        **Optional arguments:**

        *neofs*
            Number of EOFs to project onto. Defaults to all EOFs. If the
            number of EOFs requested is more than the number that are
            available, then the field will be projected onto all
            available EOFs.

        *eofscaling*
            Set the scaling of the EOFs that are projected
            onto. The following values are accepted:

            * *0* : Un-scaled EOFs (default).
            * *1* : EOFs are divided by the square-root of their eigenvalue.
            * *2* : EOFs are multiplied by the square-root of their
              eigenvalue.

        *weighted*
            If *True* then the field is weighted using the same weights
            used for the EOF analysis prior to projection. If *False*
            then no weighting is applied. Defaults to *True* (weighting
            is applied). Generally only the default setting should be
            used.

        **Returns:**

        *pseudo_pcs*
            A `~iris.cube.Cube` containing the pseudo-PCs. The PCs are
            numbered from 0 to *neofs* - 1.

        **Examples:**

        Project a field onto all EOFs::

            pseudo_pcs = solver.projectField(field)

        Project fields onto the three leading EOFs::

            pseudo_pcs = solver.projectField(field, neofs=3)

        """
        # Check that the input is an Iris cube.
        if not isinstance(cube, Cube):
            raise TypeError('the input must be an iris cube')
        cube_name = cube.name(default='dataset').replace(' ', '_')
        has_time = False
        try:
            # A time dimension must be first.
            time, time_dim = get_time_coord(cube)
            has_time = True
        except ValueError:
            # No time dimension is also acceptable.
            has_time = False
        if has_time:
            if time_dim != 0:
                raise ValueError('time must be the first dimension, '
                                 'consider using the transpose() method')
            _time_aux_coords, _, _ = classified_aux_coords(cube)
        pcs = self._solver.projectField(cube.data,
                                        neofs=neofs,
                                        eofscaling=eofscaling,
                                        weighted=weighted)
        # Create the PCs cube.
        pcs = Cube(pcs,
                   long_name='{}_pseudo_pcs'.format(cube_name),
                   var_name='pseudo_pcs')
        # Construct the required dimensions.
        if pcs.ndim == 2:
            # 2D PCs require a time axis and a PC axis.
            pcdim = DimCoord(list(range(pcs.shape[1])),
                             var_name='pc',
                             long_name='pc_number')
            pcs.add_dim_coord(copy(time), 0)
            pcs.add_dim_coord(pcdim, 1)
            # Add any time-spanning auxiliary coordinates from the input cube
            # to the returned PCs.
            for coord, dims in _time_aux_coords:
                pcs.add_aux_coord(copy(coord), dims)
        else:
            # 1D PCs require only a PC axis.
            pcdim = DimCoord(list(range(pcs.shape[0])),
                             var_name='pc',
                             long_name='pc_number')
            pcs.add_dim_coord(pcdim, 0)
        return pcs

    def getWeights(self):
        """Weights used for the analysis.

        **Returns:**

        *weights*
            An array contaning the analysis weights (not a
            `~iris.cube.Cube`).

        **Example:**

        The weights used for the analysis::

            weights = solver.getWeights()

        """
        return self._solver.getWeights()



"""Multivariate EOF analysis for `iris` cubes."""
# (c) Copyright 2013-2016 Andrew Dawson. All Rights Reserved.
#
# This file is part of eofs.
#
# eofs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# eofs is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with eofs.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import (absolute_import, division, print_function)  # noqa

import collections.abc
from copy import copy

from iris.cube import Cube
from iris.coords import DimCoord

from eofs.tools.iris import (get_time_coord, weights_array,
                             classified_aux_coords, common_items)
from . import standard


class MultivariateEof(object):
    """Multivariate EOF analysis (meta-data enabled `iris` interface)"""

    def __init__(self, cubes, weights=None, center=True, ddof=1):
        """Create a MultivariateEof instance.

        The EOF solution is computed at initialization time. Method
        calls are used to retrieve computed quantities.

        **Arguments:**

        *cubes*
            A list/tuple containing one or more `~iris.cube.Cube`
            instances, each with two or more dimensions, containing the
            data to be analysed. Time must be the first dimension of
            each `~iris.cube.Cube`. Missing values are allowed provided
            that they are constant with time in each field (e.g., values
            of an oceanographic field over land).

        **Optional arguments:**

        *weights*
            Sets the weighting method. One method can be chosen to apply
            to all cubes in *datasets* or a sequence of options can be
            given to specify a different weighting method for each cube
            in *datasets*. The following pre-defined weighting methods
            are available:

            * *'area'* : Square-root of grid cell area normalized by
              total grid area. Requires a latitude-longitude grid to be
              present in the corresponding `~iris.cube.Cube`. This is a
              fairly standard weighting strategy. If you are unsure
              which method to use and you have gridded data then this
              should be your first choice.

            * *'coslat'* : Square-root of cosine of latitude. Requires a
              latitude dimension to be present in the corresponding
              `~iris.cube.Cube`.

            * *None* : Equal weights for all grid points (*'none'* is
              also accepted).

             Alternatively a sequence of arrays of weights whose shapes
             are compatible with the corresponding `~iris.cube.Cube`
             instances in *datasets* may be supplied instead of
             specifying a weighting method.

        *center*
            If *True*, the mean along the first axis of each cube in
            *datasets* (the time-mean) will be removed prior to
            analysis. If *False*, the mean along the first axis will not
            be removed. Defaults to *True* (mean is removed).

            The covariance interpretation relies on the input data being
            anomalies with a time-mean of 0. Therefore this option
            should usually be set to *True*. Setting this option to
            *True* has the useful side effect of propagating missing
            values along the time dimension, ensuring that a solution
            can be found even if missing values occur in different
            locations at different times.

        *ddof*
            'Delta degrees of freedom'. The divisor used to normalize
            the covariance matrix is *N - ddof* where *N* is the
            number of samples. Defaults to *1*.

        **Returns:**

        *solver*
            An `MultivariateEof` instance.

        **Examples:**

        EOF analysis of two cubes with area-weighting::

            from eofs.multivariate.iris import MultivariateEof
            solver = MultivariateEof(cube1, cube2, weights='area')

        """
        # Record the number of input cubes.
        self._ncubes = len(cubes)
        # Check that the weights argument is valid and refactor it if there
        # is only one option provided.
        if weights in (None, 'area', 'coslat'):
            weights = [weights] * self._ncubes
        elif len(weights) != self._ncubes:
            raise ValueError('number of weights and cubes must match')
        # Process each input cube recording its time dimension coordinate,
        # other dimension coordinates, and defining its weight array.
        self._time = []
        self._coords = []
        self._time_aux_coords = []
        self._space_aux_coords = []
        self._time_space_aux_coords = []
        passweights = []
        for cube, weight in zip(cubes, weights):
            if not isinstance(cube, Cube):
                raise TypeError('input is not an iris cube')
            # Record the time dimension and it's position. If its position is
            # not 0 then raise an error.
            time, time_dim = get_time_coord(cube)
            if time_dim != 0:
                raise ValueError('time must be the first dimension, '
                                 'consider using the transpose() method')
            self._time.append(copy(time))
            # Make a list of the cube's other dimension coordinates.
            coords = [copy(coord) for coord in cube.dim_coords]
            coords.remove(time)
            if not coords:
                raise ValueError('one or more non-time '
                                 'dimensions are required')
            self._coords.append(coords)
            # Make a lists of the AuxCoords on the current cube and store
            # them for reapplication later.
            _t, _s, _ts = classified_aux_coords(cube)
            self._time_aux_coords.append(_t)
            self._space_aux_coords.append(_s)
            self._time_space_aux_coords.append(_ts)
            # Determine the weighting option for the cube.
            if weight is None:
                wtarray = None
            else:
                try:
                    scheme = weight.lower()
                    wtarray = weights_array(cube, scheme=scheme)
                except AttributeError:
                    wtarray = weight
            try:
                wtarray = wtarray.astype(cube.data.dtype)
            except AttributeError:
                pass
            passweights.append(wtarray)
        # Get a list of all the auxiliary coordinates that span just time
        # and are present on every input cube.
        self._common_time_aux_coords = common_items(self._time_aux_coords)
        # Create a solver.
        self._solver = standard.MultivariateEof(
            [cube.data for cube in cubes],
            weights=passweights,
            center=center,
            ddof=ddof)
        #: Number of EOFs in the solution.
        self.neofs = self._solver.neofs
        # Names of the cubes.
        self._cube_names = [c.name(default='dataset').replace(' ', '_')
                            for c in cubes]
        self._cube_var_names = [cube.var_name for cube in cubes]

    def pcs(self, pcscaling=0, npcs=None):
        """Principal component time series (PCs).

        **Optional arguments:**

        *pcscaling*
            Set the scaling of the retrieved PCs. The following
            values are accepted:

            * *0* : Un-scaled PCs (default).
            * *1* : PCs are scaled to unit variance (divided by the
              square-root of their eigenvalue).
            * *2* : PCs are multiplied by the square-root of their
              eigenvalue.

        *npcs*
            Number of PCs to retrieve. Defaults to all the PCs. If the
            number of PCs requested is more than the number that are
            available, then all available PCs will be returned.

        **Returns:**

        *pcs*
            A `~iris.cube.Cube` containing the ordered PCs.

        **Examples:**

        All un-scaled PCs::

            pcs = solver.pcs()

        First 3 PCs scaled to unit variance::

            pcs = solver.pcs(npcs=3, pcscaling=1)

        """
        pcs = self._solver.pcs(pcscaling, npcs)
        pcdim = DimCoord(list(range(pcs.shape[1])),
                         var_name='pc',
                         long_name='pc_number')
        coords = [copy(self._time[0]), pcdim]
        pcs = Cube(
            pcs,
            dim_coords_and_dims=list(zip(coords, list(range(pcs.ndim)))),
            var_name='pcs',
            long_name='principal_components')
        # Add any AuxCoords that described the time dimension of all the input
        # cubes.
        for coord, dims in self._common_time_aux_coords:
            pcs.add_aux_coord(copy(coord), dims)
        return pcs

    def eofs(self, eofscaling=0, neofs=None):
        """Empirical orthogonal functions (EOFs).

        **Optional arguments:**

        *eofscaling*
            Sets the scaling of the EOFs. The following values are
            accepted:

            * *0* : Un-scaled EOFs (default).
            * *1* : EOFs are divided by the square-root of their
              eigenvalues.
            * *2* : EOFs are multiplied by the square-root of their
              eigenvalues.

        *neofs*
            Number of EOFs to return. Defaults to all EOFs. If the
            number of EOFs requested is more than the number that are
            available, then all available EOFs will be returned.

        **Returns:**

        *eofs_list*
            A list of `~iris.cube.Cube` instances containing the ordered
            EOFs for each variable.

        **Examples:**

        All EOFs with no scaling::

            eofs_list = solver.eofs()

        The leading EOF with scaling applied::

            eof1_list = solver.eofs(neofs=1, eofscaling=1)

        """
        eofset = self._solver.eofs(eofscaling=eofscaling, neofs=neofs)
        neofs = eofset[0].shape[0]
        eofdim = DimCoord(list(range(neofs)),
                          var_name='eof',
                          long_name='eof_number')
        for iset in range(self._ncubes):
            coords = [eofdim] + [copy(coord) for coord in self._coords[iset]]
            eofset[iset] = Cube(
                eofset[iset],
                dim_coords_and_dims=list(zip(coords,
                                             range(eofset[iset].ndim))),
                var_name='eofs',
                long_name='empirical_orthogonal_functions')
            for coord, dims in self._space_aux_coords[iset]:
                eofset[iset].add_aux_coord(copy(coord), dims)
        return eofset

    def eofsAsCorrelation(self, neofs=None):
        """
        Empirical orthogonal functions (EOFs) expressed as the
        correlation between the principal component time series (PCs)
        and the each data set in the `MultivariateEof` input *datasets*.

        .. note::

            These are not related to the EOFs computed from the
            correlation matrix.

        **Optional argument:**

        *neofs*
            Number of EOFs to return. Defaults to all EOFs. If the
            number of EOFs requested is more than the number that are
            available, then all available EOFs will be returned.

        **Returns:**

        *eofs_list*
            A list of `~iris.cube.Cube` instances containing the ordered
            EOFs for each variable.

        **Examples:**

        All EOFs of each data set::

            eofs_list = solver.eofsAsCorrelation()

        The leading EOF of each data set::

            eof1_list = solver.eofsAsCorrelation(neofs=1)

        """
        eofset = self._solver.eofsAsCorrelation(neofs=neofs)
        neofs = eofset[0].shape[0]
        eofdim = DimCoord(list(range(neofs)),
                          var_name='eof',
                          long_name='eof_number')
        for iset in range(self._ncubes):
            coords = [eofdim] + [copy(coord) for coord in self._coords[iset]]
            eofset[iset] = Cube(
                eofset[iset],
                dim_coords_and_dims=list(zip(coords,
                                             range(eofset[iset].ndim))),
                var_name='eofs',
                long_name='correlation_between_pcs_and_{:s}'.format(
                          self._cube_names[iset]))
            for coord, dims in self._space_aux_coords[iset]:
                eofset[iset].add_aux_coord(copy(coord), dims)
        return eofset

    def eofsAsCovariance(self, neofs=None, pcscaling=1):
        """
        Empirical orthogonal functions (EOFs) expressed as the
        covariance between the principal component time series (PCs)
        and the each data set in the `MultivariateEof` input *datasets*.

        **Optional argument:**

        *neofs*
            Number of EOFs to return. Defaults to all EOFs. If the
            number of EOFs requested is more than the number that are
            available, then all available EOFs will be returned.

        *pcscaling*
            Set the scaling of the PCs used to compute covariance. The
            following values are accepted:

            * *0* : Un-scaled PCs.
            * *1* : PCs are scaled to unit variance (divided by the
              square-root of their eigenvalue) (default).
            * *2* : PCs are multiplied by the square-root of their
              eigenvalue.

            The default is to divide PCs by the square-root of their
            eigenvalue so that the PCs are scaled to unit variance
            (option 1).

        **Returns:**

        *eofs_list*
            A list of `~iris.cube.Cube` instances containing the ordered
            EOFs for each variable.

        **Examples:**

        All EOFs of each data set::

            eofs_list = solver.eofsAsCovariance()

        The leading EOF of each data set::

            eof1_list = solver.eofsAsCovariance(neofs=1)

        """
        eofset = self._solver.eofsAsCovariance(neofs=neofs)
        neofs = eofset[0].shape[0]
        eofdim = DimCoord(list(range(neofs)),
                          var_name='eof',
                          long_name='eof_number')
        for iset in range(self._ncubes):
            coords = [eofdim] + [copy(coord) for coord in self._coords[iset]]
            eofset[iset] = Cube(
                eofset[iset],
                dim_coords_and_dims=list(zip(coords,
                                             range(eofset[iset].ndim))),
                var_name='eofs',
                long_name='covariance_between_pcs_and_{:s}'.format(
                          self._cube_names[iset]))
            for coord, dims in self._space_aux_coords[iset]:
                eofset[iset].add_aux_coord(copy(coord), dims)
        return eofset

    def eigenvalues(self, neigs=None):
        """
        Eigenvalues (decreasing variances) associated with each EOF
        mode.

        **Optional argument:**

        *neigs*
            Number of eigenvalues to return. Defaults to all
            eigenvalues. If the number of eigenvalues requested is more
            than the number that are available, then all available
            eigenvalues will be returned.

        **Returns:**

        *eigenvalues*
            A `~iris.cube.Cube` containing the eigenvalues arranged
            largest to smallest.

        **Examples:**

        All eigenvalues::

            eigenvalues = solver.eigenvalues()

        The first eigenvalue::

            eigenvalues1 = solver.eigenvalues(neigs=1)

        """
        lambdas = self._solver.eigenvalues(neigs=neigs)
        eofdim = DimCoord(list(range(lambdas.shape[0])),
                          var_name='eigenvalue',
                          long_name='eigenvalue_number')
        coords = [eofdim]
        lambdas = Cube(
            lambdas,
            dim_coords_and_dims=list(zip(coords, range(lambdas.ndim))),
            var_name='eigenvalues',
            long_name='eigenvalues')
        return lambdas

    def varianceFraction(self, neigs=None):
        """Fractional EOF mode variances.

        **Optional argument:**

        *neigs*
            Number of eigenvalues to return the fractional variance for.
            Defaults to all eigenvalues. If the number of eigenvalues
            requested is more than the number that are available, then
            fractional variances for all available eigenvalues will be
            returned.

        **Returns:**

        *variance_fractions*
            A `~iris.cube.Cube` containing the fractional variances.

        **Examples:**

        The fractional variance represented by each EOF mode::

            variance_fractions = solver.varianceFraction()

        The fractional variance represented by the first EOF mode::

            variance_fraction_mode_1 = solver.VarianceFraction(neigs=1)

        """
        vfrac = self._solver.varianceFraction(neigs=neigs)
        eofdim = DimCoord(list(range(vfrac.shape[0])),
                          var_name='eigenvalue',
                          long_name='eigenvalue_number')
        coords = [eofdim]
        vfrac = Cube(
            vfrac,
            dim_coords_and_dims=list(zip(coords, range(vfrac.ndim))),
            var_name='variance_fraction',
            long_name='variance_fraction')
        return vfrac

    def totalAnomalyVariance(self):
        """
        Total variance associated with the field of anomalies (the sum
        of the eigenvalues).

        **Returns:**

        *total_variance*
            A scalar value (not a `~iris.cube.Cube`).

        **Example:**

        Get the total variance::

            total_variance = solver.totalAnomalyVariance()

        """
        return self._solver.totalAnomalyVariance()

    def northTest(self, neigs=None, vfscaled=False):
        """Typical errors for eigenvalues.

        The method of North et al. (1982) is used to compute the typical
        error for each eigenvalue. It is assumed that the number of
        times in the input data set is the same as the number of
        independent realizations. If this assumption is not valid then
        the result may be inappropriate.

        **Optional arguments:**

        *neigs*
            The number of eigenvalues to return typical errors for.
            Defaults to typical errors for all eigenvalues. If the
            number of eigenvalues requested is more than the number that
            are available, then typical errors for all available
            eigenvalues will be returned.

        *vfscaled*
            If *True* scale the errors by the sum of the eigenvalues.
            This yields typical errors with the same scale as the values
            returned by `MultivariateEof.varianceFraction`. If *False*
            then no scaling is done. Defaults to *False* (no scaling).

        **References**

        North G.R., T.L. Bell, R.F. Cahalan, and F.J. Moeng (1982)
        Sampling errors in the estimation of empirical orthogonal
        functions. *Mon. Weather. Rev.*, **110**, pp 669-706.

        **Returns:**

        *errors*
            A `~iris.cube.Cube` containing the typical errors.

        **Examples:**

        Typical errors for all eigenvalues::

            errs = solver.northTest()

        Typical errors for the first 5 eigenvalues scaled by the sum of
        the eigenvalues::

            errs = solver.northTest(neigs=5, vfscaled=True)

        """
        typerrs = self._solver.northTest(neigs=neigs, vfscaled=vfscaled)
        eofdim = DimCoord(list(range(typerrs.shape[0])),
                          var_name='eigenvalue',
                          long_name='eigenvalue_number')
        coords = [eofdim]
        typerrs = Cube(
            typerrs,
            dim_coords_and_dims=list(zip(coords, range(typerrs.ndim))),
            var_name='typical_errors',
            long_name='typical_errors')
        return typerrs

    def reconstructedField(self, neofs):
        """Reconstructed data sets based on a subset of EOFs.

        If weights were passed to the `MultivariateEof` instance the
        returned reconstructed fields will automatically have this
        weighting removed. Otherwise each returned field will have the
        same weighting as the corresponding  array in the
        `MultivariateEof` input *datasets*.

        **Argument:**

        *neofs*
            Number of EOFs to use for the reconstruction. If the
            number of EOFs requested is more than the number that are
            available, then all available EOFs will be used for the
            reconstruction. Alternatively this argument can be an
            iterable of mode numbers (where the first mode is 1) in
            order to facilitate reconstruction with arbitrary modes.

        **Returns:**

        *reconstruction_list*
            A list of `~iris.cube.Cube` with the same dimensions as the
            variables in the `MultivariateEof` input *datasets*
            contaning the reconstructions using *neofs* EOFs.

        **Example:**

        Reconstruct the input data sets using 3 EOFs::

            reconstruction_list = solver.reconstructedField(neofs=3)

        Reconstruct the input field using EOFs 1, 2 and 5::

            reconstruction_list = solver.reconstuctedField([1, 2, 5])

        """
        rfset = self._solver.reconstructedField(neofs)
        if isinstance(neofs, collections.abc.Iterable):
            name_part = 'EOFs_{}'.format('_'.join([str(e) for e in neofs]))
        else:
            name_part = '{:d}_EOFs'.format(neofs)
        for iset in range(self._ncubes):
            coords = [copy(self._time[iset])] + \
                     [copy(coord) for coord in self._coords[iset]]
            rfset[iset] = Cube(
                rfset[iset],
                dim_coords_and_dims=list(zip(coords, range(rfset[iset].ndim))),
                var_name=self._cube_var_names[iset] or 'dataset_{:d}'.format(
                    iset),
                long_name='{:s}_reconstructed_with_{:s}'.format(
                    self._cube_names[iset], name_part))
            rfset[iset].attributes.update({'neofs': neofs})
            for coord, dims in (self._time_aux_coords[iset] +
                                self._space_aux_coords[iset] +
                                self._time_space_aux_coords[iset]):
                rfset[iset].add_aux_coord(copy(coord), dims)
        return rfset

    def projectField(self, cubes, neofs=None, eofscaling=0, weighted=True):
        """Project a set of fields onto the EOFs.

        Given a set of fields, projects them onto the EOFs to generate a
        corresponding set of pseudo-PCs.
        **Argument:**

        *fields*
            A list/tuple containing one or more `~iris.cube.Cube`
            instances, each with two or more dimensions, containing the
            data to be projected onto the EOFs. Each field must have the
            same spatial dimensions (including missing values in the
            same places) as the corresponding data set in the
            `MultivariateEof` input *datasets*. The fields may have
            different length time dimensions to the `MultivariateEof`
            inputs *datasets* or no time dimension at all, but this
            must be consistent for all fields.

        **Optional arguments:**

        *neofs*
            Number of EOFs to project onto. Defaults to all EOFs. If the
            number of EOFs requested is more than the number that are
            available, then the field will be projected onto all
            available EOFs.

        *eofscaling*
            Set the scaling of the EOFs that are projected
            onto. The following values are accepted:

            * *0* : Un-scaled EOFs (default).
            * *1* : EOFs are divided by the square-root of their
              eigenvalue.
            * *2* : EOFs are multiplied by the square-root of their
              eigenvalue.

        *weighted*
            If *True* then each field in *fields* is weighted using the
            same weights used for the EOF analysis prior to projection.
            If *False* then no weighting is applied. Defaults to *True*
            (weighting is applied). Generally only the default setting
            should be used.

        **Returns:**

        *pseudo_pcs*
            A `~iris.cube.Cube` containing the ordered pseudo-PCs.

        **Examples:**

        Project a data set onto all EOFs::

            pseudo_pcs = solver.projectField([field1, field2])

        Project a data set onto the four leading EOFs::

            pseudo_pcs = solver.projectField([field1, field2], neofs=4)

        """
        for cube in cubes:
            if not isinstance(cube, Cube):
                raise TypeError('input is not an iris cube')
        if len(cubes) != self._ncubes:
            raise ValueError('number of cubes is incorrect, expecting {:d} '
                             'but got {:d}'.format(self._ncubes, len(cubes)))
        _all_time_aux_coords = []
        for cube in cubes:
            try:
                # Time dimension must be first.
                raise_error = False
                time, time_coord = get_time_coord(cube)
                if time_coord != 0:
                    raise_error = True
            except ValueError:
                # Not having a time dimension is also acceptable.
                pass
            if raise_error:
                raise ValueError('time must be the first dimension, '
                                 'consider using the transpose() method')
            # Store any AuxCoords describing the time dimension.
            _t, _, _ = classified_aux_coords(cube)
            _all_time_aux_coords.append(_t)
        # Retain AuxCoords that describe the time dimension of *every* input
        # cube.
        _common_time_aux_coords = common_items(_all_time_aux_coords)
        # Compute the PCs.
        pcs = self._solver.projectField([cube.data for cube in cubes],
                                        neofs=neofs,
                                        eofscaling=eofscaling,
                                        weighted=weighted)
        pcs = Cube(pcs, var_name='pseudo_pcs', long_name='pseudo_pcs')
        # Construct the required dimensions.
        if pcs.ndim == 2:
            # 2D PCs require a time axis and a PC axis.
            pcdim = DimCoord(list(range(pcs.shape[1])),
                             var_name='pc',
                             long_name='pc_number')
            time, time_dim = get_time_coord(cubes[0])
            pcs.add_dim_coord(copy(time), 0)
            pcs.add_dim_coord(pcdim, 1)
            # Add any auxiliary coordinates for the time dimension.
            for coord, dims in _common_time_aux_coords:
                pcs.add_aux_coord(copy(coord), dims)
        else:
            # 1D PCs require only a PC axis.
            pcdim = DimCoord(list(range(pcs.shape[0])),
                             var_name='pc',
                             long_name='pc_number')
            pcs.add_dim_coord(pcdim, 0)
        return pcs

    def getWeights(self):
        """Weights used for the analysis.

        **Returns:**

        *weights_list*
            A list of arrays containing the analysis weights for each
            variable (not `~iris.cube.Cube` instances).

        **Example:**

        The weights used for the analysis::

            weights_list = solver.getWeights()

        """
        return self._solver.getWeights()



"""
Compute and plot the leading EOF of sea surface temperature in the
central and northern Pacific during winter time.

The spatial pattern of this EOF is the canonical El Nino pattern, and
the associated time series shows large peaks and troughs for well-known
El Nino and La Nina events.

This example uses the metadata-retaining iris interface.

Additional requirements for this example:

    * iris (http://scitools.org.uk/iris/)
    * matplotlib (http://matplotlib.org/)
    * cartopy (http://scitools.org.uk/cartopy/)

"""
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import iris
import iris.plot as iplt
import matplotlib.pyplot as plt
import numpy as np

from eofs.iris import Eof
from eofs.examples import example_data_path


# Read SST anomalies using the iris module. The file contains November-March
# averages of SST anomaly in the central and northern Pacific.
filename = example_data_path('sst_ndjfm_anom.nc')
sst = iris.load_cube(filename)

# Create an EOF solver to do the EOF analysis. Square-root of cosine of
# latitude weights are applied before the computation of EOFs.
solver = Eof(sst, weights='coslat')

# Retrieve the leading EOF, expressed as the correlation between the leading
# PC time series and the input SST anomalies at each grid point, and the
# leading PC time series itself.
eof1 = solver.eofsAsCorrelation(neofs=1)
pc1 = solver.pcs(npcs=1, pcscaling=1)

# Plot the leading EOF expressed as correlation in the Pacific domain.
clevs = np.linspace(-1, 1, 11)
ax = plt.axes(projection=ccrs.PlateCarree(central_longitude=190))
fill = iplt.contourf(eof1[0], clevs, cmap=plt.cm.RdBu_r)
ax.add_feature(cfeature.LAND, facecolor='w', edgecolor='k')
cb = plt.colorbar(fill, orientation='horizontal')
cb.set_label('correlation coefficient', fontsize=12)
ax.set_title('EOF1 expressed as correlation', fontsize=16)

# Plot the leading PC time series.
plt.figure()
iplt.plot(pc1[:, 0], color='b', linewidth=2)
ax = plt.gca()
ax.axhline(0, color='k')
ax.set_ylim(-3, 3)
ax.set_xlabel('Year')
ax.set_ylabel('Normalized Units')
ax.set_title('PC1 Time Series', fontsize=16)

plt.show()



"""Supplementary tools useful for multiple interfaces."""
# (c) Copyright 2013-2016 Andrew Dawson. All Rights Reserved.
#
# This file is part of eofs.
#
# eofs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# eofs is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with eofs.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import (absolute_import, division, print_function)  # noqa


def covcor_dimensions(pc_dims, field_dims):
    """
    Extract the appropriate dimensions from a set of PCs and a field for
    construction of covariance/correlation map dimensions.

    """
    spatial_dims = field_dims[1:]
    try:
        pc_dim = pc_dims[1]
    except IndexError:
        pc_dim = None
    covcor_dims = [d for d in [pc_dim] + spatial_dims if d is not None]
    return covcor_dims



"""Utilities for testing the `eofs` package."""
# (c) Copyright 2013-2016 Andrew Dawson. All Rights Reserved.
#
# This file is part of eofs.
#
# eofs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# eofs is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with eofs.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import (absolute_import, division, print_function)  # noqa

import numpy as np
try:
    from iris.cube import Cube
except ImportError:
    pass


def _close(a, b, rtol=1e-05, atol=1e-08):
    """Check if two values are close."""
    return abs(a - b) <= (atol + rtol * abs(b))


def __tomasked(*args):
    """Convert cdms2 variables or iris cubes to masked arrays.

    The conversion is safe, so if non-variables/cubes are passed they
    are just returned.

    """
    def __asma(a):
        try:
            if type(a) is Cube:
                # Retrieve the data from the cube.
                a = a.data
        except NameError:
            pass
        try:
            # Retrieve data from cdms variable.
            a = a.asma()
        except AttributeError:
            # The input is already an array or masked array, either extracted
            # from an iris cube, or was like that to begin with.
            pass
        return a
    return [__asma(a) for a in args]


def error(a, b):
    """Compute the error between two arrays.

    Computes the RMSD normalized by the range of the second input.

    """
    a, b = __tomasked(a, b)
    return np.sqrt((a - b)**2).mean() / (np.max(b) - np.min(b))


def sign_adjustments(eofset, refeofset):
    """Sign adjustments for EOFs/PCs.

    Create a matrix of sign weights used for adjusting the sign of a set
    of EOFs or PCs to the sign of a reference set.

    The first dimension is assumed to be modes.

    **Arguments:**

    *eofset*
        Set of EOFs.

    *refeofset*
        Reference set of EOFs.

    """
    if eofset.shape != refeofset.shape:
        raise ValueError('input set has different shape from reference set')
    eofset, refeofset = __tomasked(eofset, refeofset)
    nmodes = eofset.shape[0]
    signs = np.ones([nmodes])
    shape = [nmodes] + [1] * (eofset.ndim - 1)
    eofset = eofset.reshape([nmodes, np.prod(eofset.shape[1:], dtype=int)])
    refeofset = refeofset.reshape([nmodes,
                                   np.prod(refeofset.shape[1:],
                                           dtype=int)])
    for mode in range(nmodes):
        i = 0
        try:
            while _close(eofset[mode, i], 0.) or \
                  _close(refeofset[mode, i], 0.) \
                  or np.ma.is_masked(eofset[mode, i]) or \
                  np.ma.is_masked(refeofset[mode, i]):
                i += 1
        except IndexError:
            i = 0
        if np.sign(eofset[mode, i]) != np.sign(refeofset[mode, i]):
            signs[mode] = -1
    return signs.reshape(shape)


if __name__ == '__main__':
    pass



"""Supplementary tools for the `numpy` EOF analysis interface."""
# (c) Copyright 2010-2015 Andrew Dawson. All Rights Reserved.
#
# This file is part of eofs.
#
# eofs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# eofs is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with eofs.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import (absolute_import, division, print_function)  # noqa

import numpy as np
import numpy.ma as ma


def _check_flat_center(pcs, field):
    """
    Check PCs and a field for shape compatibility, flatten both to 2D,
    and center along the first dimension.

    This set of operations is common to both covariance and correlation
    calculations.

    """
    # Get the number of times in the field.
    records = field.shape[0]
    if records != pcs.shape[0]:
        # Raise an error if the field has a different number of times to the
        # PCs provided.
        raise ValueError("PCs and field must have the same first dimension")
    if len(pcs.shape) > 2:
        # Raise an error if the PCs are more than 2D.
        raise ValueError("PCs must be 1D or 2D")
    # Check if the field is 1D.
    if len(field.shape) == 1:
        originalshape = tuple()
        channels = 1
    else:
        # Record the shape of the field and the number of spatial elements.
        originalshape = field.shape[1:]
        channels = np.product(originalshape)
    # Record the number of PCs.
    if len(pcs.shape) == 1:
        npcs = 1
        npcs_out = tuple()
    else:
        npcs = pcs.shape[1]
        npcs_out = (npcs,)
    # Create a flattened field so iterating over space is simple. Also do this
    # for the PCs to ensure they are 2D.
    field_flat = field.reshape([records, channels])
    pcs_flat = pcs.reshape([records, npcs])
    # Centre both the field and PCs in the time dimension.
    field_flat = field_flat - field_flat.mean(axis=0)
    pcs_flat = pcs_flat - pcs_flat.mean(axis=0)
    return pcs_flat, field_flat, npcs_out + originalshape


def correlation_map(pcs, field):
    """Correlation maps for a set of PCs and a spatial-temporal field.

    Given an array where the columns are PCs (e.g., as output from
    `~eofs.standard.Eof.pcs`) and an array containing spatial-temporal
    data where the first dimension represents time, one correlation map
    per PC is computed.

    The field must have the same temporal dimension as the PCs. Any
    number of spatial dimensions (including zero) are allowed in the
    field and there can be any number of PCs.

    **Arguments:**

    *pcs*
        PCs as the columns of an array.

    *field*
        Spatial-temporal field with time as the first dimension.

    **Returns:**

    *correlation_maps*
        An array with the correlation maps along the first dimension.

    **Example:**

    Compute correlation maps for each PC::

        pcs = solver.pcs(pcscaling=1)
        correlation_maps = correlation_maps(pcs, field)

    """
    # Check PCs and fields for validity, flatten the arrays ready for the
    # computation and remove the mean along the leading dimension.
    pcs_cent, field_cent, out_shape = _check_flat_center(pcs, field)
    # Compute the standard deviation of the PCs and the fields along the time
    # dimension (the leading dimension).
    pcs_std = pcs_cent.std(axis=0)
    field_std = field_cent.std(axis=0)
    # Set the divisor.
    div = np.float64(pcs_cent.shape[0])
    # Compute the correlation map.
    cor = ma.dot(field_cent.T, pcs_cent).T / div
    cor = ma.masked_invalid(cor)
    cor /= ma.outer(pcs_std, field_std)
    # Return the correlation with the appropriate shape.
    return cor.reshape(out_shape)


def covariance_map(pcs, field, ddof=1):
    """Covariance maps for a set of PCs and a spatial-temporal field.

    Given an array where the columns are PCs (e.g., as output from
    `eofs.standard.Eof.pcs`) and an array containing spatial-temporal
    data where the first dimension represents time, one covariance map
    per PC is computed.

    The field must have the same temporal dimension as the PCs. Any
    number of spatial dimensions (including zero) are allowed in the
    field and there can be any number of PCs.

    **Arguments:**

    *pcs*
        PCs as the columns of an array.

    *field*
        Spatial-temporal field with time as the first dimension.

    **Optional arguments:**

    *ddof*
        'Delta degrees of freedom'. The divisor used to normalize
        the covariance matrix is *N - ddof* where *N* is the
        number of samples. Defaults to *1*.

    **Returns:**

    *covariance_maps*
        An array with the covariance maps along the first dimension.

    **Example:**

    Compute covariance maps for each PC::

        pcs = solver.pcs(pcscaling=1)
        covariance_maps = covariance_maps(pcs, field)

    """
    # Check PCs and fields for validity, flatten the arrays ready for the
    # computation and remove the mean along the leading dimension.
    pcs_cent, field_cent, out_shape = _check_flat_center(pcs, field)
    # Set the divisor according to the specified delta-degrees of freedom.
    div = np.float64(pcs_cent.shape[0] - ddof)
    # Compute the covariance map, making sure it has the appropriate shape.
    cov = (ma.dot(field_cent.T, pcs_cent).T / div).reshape(out_shape)
    cov = ma.masked_invalid(cov)
    return cov



"""Supplementary tools for the `cdms` EOF analysis interface."""
# (c) Copyright 2010-2013 Andrew Dawson. All Rights Reserved.
#
# This file is part of eofs.
#
# eofs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# eofs is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with eofs.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import (absolute_import, division, print_function)  # noqa

import cdms2
import numpy as np

from .standard import covariance_map as standard_covmap
from .standard import correlation_map as standard_cormap
from .generic import covcor_dimensions


def _rootcoslat_weights(latdim):
    """Square-root of cosine of latitude weights.

    *latdim*
       Latitude dimension values.

    """
    coslat = np.cos(np.deg2rad(latdim))
    coslat[np.where(coslat < 0)] = 0.
    latw = np.sqrt(coslat)
    latw[np.where(np.isnan(latw))] = 0.
    return latw


def _area_weights(grid, gridorder):
    """Area weights.

    *grid*
        `cdms2` grid.

    *gridorder*
        Either "xy" or "yx".

    """
    latw, lonw = grid.getWeights()
    if gridorder == "xy":
        wtarray = np.outer(lonw, latw)
    else:
        wtarray = np.outer(latw, lonw)
    wtarray /= wtarray.sum()
    wtarray = np.sqrt(wtarray)
    return wtarray


def weights_array(dataset, scheme):
    """Weights for a data set on a grid.

    Returned weights are a `numpy.ndarray` broadcastable to the shape of
    the input data.

    **Arguments:**

    *dataset*
        A `cdms2` variable to generate weights for.

    *scheme*
        Weighting scheme to use. The following values are accepted:

        * *'coslat'* : Square-root of cosine of latitude.
        * *'area'* : Square-root of grid cell area normalized by total
                     grid area.

    **Returns:**

    *weights*
        An array contanining the weights (not a `cdms2` variable).

    **Examples:**

    Area weights for a `cdms2` variable on 2D grid:

        weights = weights_array(var2d, scheme='area')

    Square-root of cosine of latitude weights for a `cdms2` variable
    with a latitude dimension:

        weights = weights_array(var, scheme='coslat')

    """
    # A re-usable generic error message for the function. When raising an
    # exception just fill in what is required.
    errstr = "weighting scheme '{!s}' requires {{!s}}".format(scheme)
    # Always use lower-case for the scheme, allowing the user to use
    # upper-case in their calling code without an error.
    scheme = scheme.lower()
    if scheme in ('area'):
        # Handle area weighting.
        grid = dataset.getGrid()
        if grid is None:
            raise ValueError(errstr.format('a grid'))
        order = dataset.getOrder()
        if 'xy' in order:
            gridorder = 'xy'
            dimtoindex = dataset.getLatitude()
        elif 'yx' in order:
            gridorder = 'yx'
            dimtoindex = dataset.getLongitude()
        else:
            raise ValueError(
                errstr.format('adjacent latitude and longitude dimensions'))
        # Retrieve area weights for the specified grid.
        weights = _area_weights(grid, gridorder)
    elif scheme in ('coslat', 'cos_lat'):
        # Handle square-root of cosine of latitude weighting.
        try:
            latdim = dataset.getLatitude()[:]
            dimtoindex = dataset.getLatitude()
        except (AttributeError, TypeError):
            raise ValueError(errstr.format('a latitude dimension'))
        # Retrieve latitude weights.
        weights = _rootcoslat_weights(latdim)
    else:
        raise ValueError("invalid weighting scheme: '{!s}'".format(scheme))
    # Re-shape the retrieved weights so that they are broadcastable to the
    # shape of the input arrays. This just involves adding any additional
    # singleton dimensions to the right of the last weights dimension.
    rightdims = len(dataset.shape) - dataset.getAxisIndex(dimtoindex) - 1
    weights = weights.reshape(weights.shape + (1,) * rightdims)
    return weights


def cdms2_name(variable):
    """Return a human-readable name of a cdms2 variable.

    First it tries 'standard_name', then 'long_name' then 'id'.

    """
    names = [getattr(variable, 'standard_name', None),
             getattr(variable, 'long_name', None),
             getattr(variable, 'id', None)]
    try:
        return [name for name in names if name is not None][0]
    except IndexError:
        return 'dataset'


def correlation_map(pcs, field):
    """Correlation maps for a set of PCs and a spatial-temporal field.

    Given a set of PCs in a `cdms2` variable (e.g., as output from
    `eofs.cdms.Eof.pcs`) and a spatial-temporal field in a `cdms`
    variable, one correlation map per PC is computed.

    The field must have the same temporal dimension as the PCs. Any
    number of spatial dimensions (including zero) are allowed in the
    field and there can be any number of PCs.

    **Arguments:**

    *pcs*
        PCs in a `cdms2` variable.

    *field*
        Spatial-temporal field in a `cdms2` variable.

    **Returns:**

    *correlation_maps*
        A `cdms2` variable containing the correlation maps.

    **Examples:**

    Compute correlation maps for each PC::

        pcs = eofobj.pcs(pcscaling=1)
        cormaps = correlation_map(pcs, field)

    """
    cor = standard_cormap(pcs.asma(), field.asma())
    outdims = covcor_dimensions(pcs.getAxisList(), field.getAxisList())
    if not outdims:
        # There are no output dimensions, return a scalar.
        return cor
    # Otherwise return a cdms2 variable with the appropriate dimensions.
    cor = cdms2.createVariable(cor, axes=outdims, id='pccor')
    cor.long_name = "pc_correlation"
    return cor


def covariance_map(pcs, field, ddof=1):
    """Covariance maps for a set of PCs and a spatial-temporal field.

    Given a set of PCs in a `cdms2` variable (e.g., as output from
    `~eofs.cdms.Eof.pcs`) and a spatial-temporal field in a `cdms2`
    variable, one covariance map per PC is computed.

    The field must have the same temporal dimension as the PCs. Any
    number of spatial dimensions (including zero) are allowed in the
    field and there can be any number of PCs.

    **Arguments:**

    *pcs*
        PCs in a `cdms2` variable.

    *field*
        Spatial-temporal field in a `cdms2` variable.

    **Optional arguments:**

    *ddof*
        'Delta degrees of freedom'. The divisor used to normalize
        the covariance matrix is *N - ddof* where *N* is the
        number of samples. Defaults to *1*.

    **Returns:**

    *covariance_maps*
        A `cdms2` variable containing the covariance maps.

    **Examples:**

    Compute covariance maps for each PC::

        pcs = eofobj.pcs(pcscaling=1)
        covmaps = covariance_map(pcs, field)

    """
    cov = standard_covmap(pcs.asma(), field.asma(), ddof=ddof)
    outdims = covcor_dimensions(pcs.getAxisList(), field.getAxisList())
    if not outdims:
        # There are no output dimensions, return a scalar.
        return cov
    # Otherwise return a cdms2 variable with the appropriate dimensions.
    cov = cdms2.createVariable(cov, axes=outdims, id='pccov')
    cov.long_name = 'pc_covariance'
    return cov



"""Supplementary tools for the `iris` EOF analysis interface."""
# (c) Copyright 2013-2016 Andrew Dawson. All Rights Reserved.
#
# This file is part of eofs.
#
# eofs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# eofs is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with eofs.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import (absolute_import, division, print_function)  # noqa
from copy import copy
from functools import reduce
import warnings

import numpy as np
from iris.cube import Cube
from iris.analysis.cartography import area_weights
from iris.analysis.cartography import cosine_latitude_weights
from iris.exceptions import CoordinateMultiDimError, CoordinateNotFoundError

from . import standard
from .generic import covcor_dimensions


def weights_array(cube, scheme):
    """Weights for a data set on a grid.

    Returned weights are a `numpy.ndarray` broadcastable to the shape of
    the input cube.

    **Arguments:**

    *cube*
        An `~iris.cube.Cube` instance to generate weights for.

    *scheme*
        Weighting scheme to use. The following values are accepted:

        * *'coslat'* : Square-root of cosine of latitude.
        * *'area'* : Square-root of grid cell area normalized by total
                     grid area.

    **Returns:**

    *weights*
        An array contanining the weights (not a `~iris.cube.Cube`).

    **Examples:**

    Area weights for a `~iris.cube.Cube` on 2D grid:

        weights = weights_array(cube, scheme='area')

    Square-root of cosine of latitude weights for a `~iris.cube.Cube`
    with a latitude dimension:

        weights = weights_array(cube, scheme='coslat')

    """
    # Always use lower-case for the scheme, allowing the user to use
    # upper-case in their calling code without an error.
    scheme = scheme.lower()
    if scheme in ('area',):
        # Handle area weighting.
        try:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                weights = np.sqrt(area_weights(cube, normalize=True))
        except (ValueError, CoordinateMultiDimError):
            raise ValueError('cannot generate area weights')
    elif scheme in ('coslat',):
        # Handle square-root of cosine of latitude weighting.
        try:
            weights = np.sqrt(cosine_latitude_weights(cube))
        except (ValueError, CoordinateMultiDimError):
            raise ValueError('cannot generate latitude weights')
    else:
        raise ValueError("invalid weighting scheme: '{!s}'".format(scheme))
    return weights


def coord_and_dim(cube, coord, multiple=False):
    """
    Retrieve a coordinate dimension and its corresponding position from
    a `~iris.cube.Cube` instance.

    **Arguments:**

    *cube*
        An `~iris.cube.Cube` instance to retrieve the dimension from.

    *coord*
        Name of the coordinate dimension to retrieve.

    **Returns:**

    *coord_tuple*
        A 2-tuple of (coordinate_dimension, dimension_number).

    .. deprecated:: version 1.2

       The function `get_time_coord` is used instead for finding time
       coordinates. For other coordinates please use the iris built-in
       functionality for locating required cooridnates.

    """
    deprecation_message = ('coord_and_dim() is deprecated, please use '
                           'get_time_coord() or built-in iris functionality')
    warnings.warn(deprecation_message, DeprecationWarning)
    coords = [c for c in cube.dim_coords if coord in c.name()]
    if len(coords) > 1:
        raise ValueError('multiple {} coordinates are not '
                         'allowed'.format(coord))
    try:
        c = coords[0]
    except IndexError:
        raise ValueError('cannot get {!s} coordinate from '
                         'cube {!r}'.format(coord, cube))
    c_dim = cube.coord_dims(c)
    c_dim = c_dim[0] if c_dim else None
    return c, c_dim


def get_time_coord(cube):
    """
    Retrieve the time coordinate dimension and its corresponding
    position from a `~iris.cube.Cube` instance.

    **Arguments:**

    *cube*
        An `~iris.cube.Cube` instance to retrieve the dimension from.

    **Returns:**

    *coord_tuple*
        A 2-tuple of (coordinate_dimension, dimension_number).

    """
    time_coords = cube.coords(axis='T', dim_coords=True)
    if not time_coords:
        # If no coordinates were identified as time, relax the criteria and
        # look for dimension coordinates with 'time' in the name:
        time_coords = [coord for coord in cube.dim_coords
                       if 'time' in coord.name()]
        if not time_coords:
            raise ValueError('cannot find a time dimension coordinate')
    if len(time_coords) > 1:
        raise ValueError('multiple time coordinates are not allowed')
    time_coord = time_coords[0]
    time_dim = cube.coord_dims(time_coord)[0]
    return time_coord, time_dim


def classified_aux_coords(cube):
    """
    Classify a Cube's auxiliary coordinates into those that span only
    the time dimension, those that span only space dimensions, and those
    that span both time and space dimensions.

    **Arguments:**

    *cube*
        An `~iris.cube.Cube` instance whose auxiliary coordinates should
        be classified.

    **Returns:**

    *coord_sets*
        A 3-tuple of lists of coordinates. The first element is the list
        of all auxiliary coordinates spannning only the time dimension,
        the second element is the list of all auxiliary coordinates
        spannning only space dimensions, and the third element is the
        list of all auxiliary coordinates spannning both time and space
        dimensions.

    """
    try:
        _, timedim = get_time_coord(cube)
    except ValueError:
        timedim = None
    time_only = []
    space_only = []
    time_and_space = []
    for coord in cube.aux_coords:
        dims = cube.coord_dims(coord)
        if dims == (timedim,):
            time_only.append((copy(coord), timedim))
        elif dims:
            if timedim in dims:
                time_and_space.append((copy(coord), dims))
            else:
                space_only.append((copy(coord), dims))
    return time_only, space_only, time_and_space


def common_items(item_set):
    """
    Given an iterable of lists, constructs a list of every item that is
    present in all of the lists.

    **Arguments:**

    *item_set*
        An iterable containing lists of items.

    **Returns:**

    *common*
        A list of the items which occur in all sublists in the input.

    """
    common = []
    for item in reduce(lambda x, y: x + y, item_set):
        item_ok = all([item in items for items in item_set])
        if item_ok and item not in common:
            common.append(item)
    return common


def _time_coord_info(cube):
    time, time_dim = get_time_coord(cube)
    coords = [copy(coord) for coord in cube.dim_coords]
    coords.remove(time)
    coords = [copy(time)] + coords
    return time_dim, coords


def _map_and_dims(pcs, field, mapfunc, *args, **kwargs):
    """
    Compute a set of covariance/correlation maps and the resulting
    dimensions.

    """
    info = {}
    for cube in (field, pcs):
        info[cube.name()] = _time_coord_info(cube)
    cmap_args = [np.rollaxis(cube.data, info[cube.name()][0])
                 for cube in (pcs, field)]
    cmap_args += args
    dim_args = [info[cube.name()][1] for cube in (pcs, field)]
    cmap = mapfunc(*cmap_args, **kwargs)
    dims = covcor_dimensions(*dim_args)
    return cmap, dims


def correlation_map(pcs, field):
    """Correlation between PCs and a field.

    Computes maps of the correlation between each PC and the given
    field at each grid point.

    Given a set of PCs contained in a `~iris.cube.Cube` (e.g., as output
    from `eofs.iris.Eof.pcs`) and a field with a time dimension
    contained in a `iris.cube.Cube`, one correlation map per PC is
    computed.

    The field must have the same length time dimension as the PCs. Any
    number of spatial dimensions (including zero) are allowed in the
    field and there can be any number of PCs.

    **Arguments:**

    *pcs*
        PCs contained in a `~iris.cube.Cube`.

    *field*
        Field Spatial-temporal field contained in a `~iris.cube.Cube`.

    **Returns:**

    *correlation_maps*
        A `~iris.cube.Cube` containing the correlation maps.

    **Examples:**

    Assuming *solver* is an instance of `eofs.iris.Eof`, compute
    correlation maps for each PC::

        pcs = solver.pcs(pcscaling=1)
        correlation_maps = correlation_map(pcs, field)

    """
    # Compute the correlation map and retrieve appropriate Iris coordinate
    # dimensions for it.
    cor, dims = _map_and_dims(pcs, field, standard.correlation_map)
    if not dims:
        # There are no output dimensions, return a scalar.
        return cor
    # Otherwise return an Iris cube.
    cor = Cube(cor, dim_coords_and_dims=list(zip(dims, range(cor.ndim))))
    cor.long_name = 'pc_correlation'
    return cor


def covariance_map(pcs, field, ddof=1):
    """Covariance between PCs and a field.

    Computes maps of the covariance between each PC and the given
    field at each grid point.

    Given a set of PCs contained in a `~iris.cube.Cube` (e.g., as output
    from `eofs.iris.Eof.pcs`) and a field with a time dimension
    contained in a `iris.cube.Cube`, one covariance map per PC is
    computed.

    The field must have the same length time dimension as the PCs. Any
    number of spatial dimensions (including zero) are allowed in the
    field and there can be any number of PCs.

    **Arguments:**

    *pcs*
        PCs contained in a `~iris.cube.Cube`.

    *field*
        Field Spatial-temporal field contained in a `~iris.cube.Cube`.

    **Keyword arguments:**

    *ddof*
        'Delta degrees of freedom'. The divisor used to normalize
        the covariance matrix is *N - ddof* where *N* is the
        number of samples. Defaults to *1*.

    **Returns:**

    *covariance_maps*
        A `~iris.cube.Cube` containing the covariance maps.

    **Examples:**

    Compute covariance maps for each PC::

        pcs = solver.pcs(pcscaling=1)
        covariance_maps = covariance_map(pcs, field)

    """
    # Compute the covariance map and retrieve appropriate Iris coordinate
    # dimensions for it.
    cov, dims = _map_and_dims(pcs, field, standard.covariance_map, ddof=ddof)
    if not dims:
        # There are no output dimensions, return a scalar.
        return cov
    # Otherwise return an Iris cube.
    cov = Cube(cov, dim_coords_and_dims=list(zip(dims, range(cov.ndim))))
    cov.long_name = 'pc_covariance'
    return cov



"""Tests for error handling in `eofs`."""
# (c) Copyright 2013-2016 Andrew Dawson. All Rights Reserved.
#
# This file is part of eofs.
#
# eofs is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# eofs is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with eofs.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import (absolute_import, division, print_function)  # noqa

import numpy as np
try:
    import cdms2
except ImportError:
    pass
import pytest

import eofs
from eofs.tests import EofsTest

from .reference import reference_solution


# Create a mapping from interface name to solver class.
solvers = {'standard': eofs.standard.Eof}
try:
    solvers['cdms'] = eofs.cdms.Eof
except AttributeError:
    pass
try:
    solvers['iris'] = eofs.iris.Eof
except AttributeError:
    pass
try:
    solvers['xarray'] = eofs.xarray.Eof
except AttributeError:
    pass


class ErrorHandlersTest(EofsTest):
    interface = None
    weights = None

    @classmethod
    def setup_class(cls):
        try:
            cls.solution = reference_solution(cls.interface, cls.weights)
        except ValueError:
            pytest.skip('missing dependencies required to test '
                        'the {!s} interface'.format(cls.interface))
        cls.neofs = cls.solution['eigenvalues'].shape[0]
        try:
            cls.solver = solvers[cls.interface](
                cls.solution['sst'], weights=cls.solution['weights'])
        except KeyError:
            pytest.skip('missing dependencies required to test '
                        'the {!s} interface'.format(cls.interface))

    def test_invalid_pcscaling(self):
        # PCs with invalid scaling
        with pytest.raises(ValueError):
            pcs = self.solver.pcs(pcscaling=-1)

    def test_invalid_eofscaling(self):
        # EOFs with invalid scaling
        with pytest.raises(ValueError):
            eofs = self.solver.eofs(eofscaling=-1)

    def test_projectField_invalid_dimensions(self):
        # projecting a field with too few dimensions
        data = self.solution['sst'][:, 0, 0]
        with pytest.raises(ValueError):
            pcs = self.solver.projectField(data)

    def test_projectField_invalid_shape(self):
        # projecting a field with the wrong shape
        data = self.solution['sst'][..., 0:1]
        with pytest.raises(ValueError):
            pcs = self.solver.projectField(data)

    def test_projectField_different_missing_values(self):
        # projecting a field with different missing values
        solution = reference_solution(self.interface, self.weights)
        data = solution['sst']
        mask = data.mask
        mask[0] = True
        with pytest.raises(ValueError):
            pcs = self.solver.projectField(data)


# ----------------------------------------------------------------------------
# Error Handler tests for the standard interface


class TestErroHandlersStandard(ErrorHandlersTest):
    interface = 'standard'
    weights = 'equal'


# ----------------------------------------------------------------------------
# Error Handler tests for the cdms interface


class TestErrorHandlersCDMS(ErrorHandlersTest):
    """Test error handling in the cdms interface."""
    interface = 'cdms'
    weights = 'equal'

    def test_projectField_invalid_type(self):
        # projecting a field of the wrong type
        solution = reference_solution('standard', 'equal')
        with pytest.raises(TypeError):
            pcs = self.solver.projectField(solution['sst'])


# ----------------------------------------------------------------------------
# Error Handler tests for the iris interface


class TestErrorHandlersIris(ErrorHandlersTest):
    """Test error handling in the iris interface."""
    interface = 'iris'
    weights = 'equal'

    def test_projectField_invalid_type(self):
        # projecting a field of the wrong type
        solution = reference_solution('standard', 'equal')
        with pytest.raises(TypeError):
            pcs = self.solver.projectField(solution['sst'])

    def test_projectField_different_missing_values(self):
        # projecting a field with different missing values
        solution = reference_solution(self.interface, self.weights)
        data = solution['sst']
        mask = data.data.mask
        mask[0] = True
        with pytest.raises(ValueError):
            pcs = self.solver.projectField(data)

    def test_projectField_invalid_dimension_order(self):
        # projecting a field with the time dimension not at the front
        solution = reference_solution(self.interface, self.weights)
        data = solution['sst']
        data.transpose((2, 1, 0))
        with pytest.raises(ValueError):
            pcs = self.solver.projectField(data)


# ----------------------------------------------------------------------------
# Error Handler tests for the xarray interface


class TestErrorHandlersXarray(ErrorHandlersTest):
    """Test error handling in the xarray interface."""
    interface = 'xarray'
    weights = 'equal'

    def test_projectField_invalid_type(self):
        # projecting a field of the wrong type
        solution = reference_solution('standard', 'equal')
        with pytest.raises(TypeError):
            pcs = self.solver.projectField(solution['sst'])

    def test_projectField_different_missing_values(self):
        # projecting a field with different missing values
        solution = reference_solution(self.interface, self.weights)
        data = solution['sst']
        data.values[0] = np.nan
        with pytest.raises(ValueError):
            pcs = self.solver.projectField(data)

    def test_projectField_invalid_dimension_order(self):
        # projecting a field with the time dimension not at the front
        solution = reference_solution(self.interface, self.weights)
        data = solution['sst']
        data = data.transpose('longitude', 'latitude', 'time')
        with pytest.raises(ValueError):
            pcs = self.solver.projectField(data)


# ----------------------------------------------------------------------------
# Constructor tests for the standard interface


class TestConstructorStandard(EofsTest):
    """Test the error handling in the standard interface constructor."""

    @classmethod
    def setup_class(cls):
        cls.solver_class = solvers['standard']

    def test_invalid_input_dimensions(self):
        # too few input dimensions
        solution = reference_solution('standard', 'equal')
        data = solution['sst'][:, 0, 0]
        with pytest.raises(ValueError):
            solver = self.solver_class(data)

    def test_invalid_weights_dimensions(self):
        # weights with incompatible dimensions
        solution = reference_solution('standard', 'area')
        data = solution['sst']
        weights = solution['weights'][:, 0]
        with pytest.raises(ValueError):
            solver = self.solver_class(data, weights=weights)

    def test_invalid_weights_type(self):
        # weights with an incompatible type
        solution = reference_solution('standard', 'area')
        data = solution['sst']
        weights = 'area'
        with pytest.raises(TypeError):
            solver = self.solver_class(data, weights=weights)

    def test_input_with_all_missing_values(self):
        # input with only missing values, missing values are propagated
        # because the default for the center argument is True
        solution = reference_solution('standard', 'equal')
        data = solution['sst']
        mask = data.mask
        mask[-1] = True
        with pytest.raises(ValueError):
            solver = self.solver_class(data)

    def test_input_with_non_uniform_missing_values(self):
        # missing values in different places at different times leads to
        # errors in computing the SVD
        solution = reference_solution('standard', 'equal')
        data = solution['sst']
        mask = data.mask
        mask[-1] = True
        with pytest.raises(ValueError):
            solver = self.solver_class(data, center=False)


# ----------------------------------------------------------------------------
# Constructor tests for the cdms interface


class TestConstructorCDMS(EofsTest):
    """Test the error handling in the cdms interface constructor."""

    @classmethod
    def setup_class(cls):
        try:
            cls.solver_class = solvers['cdms']
        except KeyError:
            pytest.skip('missing dependencies required to test the '
                        'cdms interface')

    def test_wrong_input_type(self):
        # input of the wrong type
        solution = reference_solution('standard', 'equal')
        data = solution['sst']
        with pytest.raises(TypeError):
            solver = self.solver_class(data)

    def test_input_without_time_dimension(self):
        # no time dimension in the input
        solution = reference_solution('cdms', 'equal')
        data = solution['sst'](time=slice(0, 1), squeeze=True)
        with pytest.raises(ValueError):
            solver = self.solver_class(data)

    def test_input_time_dimension_not_first(self):
        # time not the first dimension in the input
        solution = reference_solution('cdms', 'equal')
        data = solution['sst']
        data = data.reorder('xyt')
        with pytest.raises(ValueError):
            solver = self.solver_class(data)

    def test_input_no_spatial_dimensions(self):
        # not enough dimensions in the input
        solution = reference_solution('cdms', 'equal')
        data = solution['sst'][:, 0, 0]
        with pytest.raises(ValueError):
            solver = self.solver_class(data)

    def test_invalid_builtin_weights_value(self):
        # invalid weighting scheme name
        solution = reference_solution('cdms', 'equal')
        data = solution['sst']
        with pytest.raises(ValueError):
            solver = self.solver_class(data, weights='invalid')

    def test_builtin_latitude_weights_with_missing_dimension(self):
        # latitude weights without latitude dimension
        solution = reference_solution('cdms', 'equal')
        data = solution['sst'](latitude=slice(0, 1), squeeze=True)
        with pytest.raises(ValueError):
            solver = self.solver_class(data, weights='coslat')

    def test_builtin_area_weights_with_missing_dimension(self):
        # area weights without longitude dimension
        solution = reference_solution('cdms', 'equal')
        data = solution['sst'](longitude=slice(0, 1), squeeze=True)
        with pytest.raises(ValueError):
            solver = self.solver_class(data, weights='area')

    def test_builtin_area_weights_with_non_adjacent_dimensions(self):
        # area weights with latitude and longitude not adjacent in input
        solution = reference_solution('cdms', 'equal')
        data = solution['sst']
        newdim = cdms2.createAxis([0.], id='height')
        newdim.designateLevel()
        data = cdms2.createVariable(cdms2.MV.reshape(data, data.shape + (1,)),
                                    axes=data.getAxisList() + [newdim],
                                    id=data.id)
        data = data.reorder('txzy')
        with pytest.raises(ValueError):
            solver = self.solver_class(data, weights='area')


# ----------------------------------------------------------------------------
# Constructor tests for the iris interface


class TestConstructorIris(EofsTest):
    """Test the error handling in the iris interface constructor."""

    @classmethod
    def setup_class(cls):
        try:
            cls.solver_class = solvers['iris']
        except KeyError:
            pytest.skip('missing dependencies required to test the '
                        'iris interface')

    def test_wrong_input_type(self):
        # input of the wrong type
        solution = reference_solution('standard', 'equal')
        data = solution['sst']
        with pytest.raises(TypeError):
            solver = self.solver_class(data)

    def test_input_without_time_dimension(self):
        # no time dimension in the input
        solution = reference_solution('iris', 'equal')
        data = solution['sst'][0]
        with pytest.raises(ValueError):
            solver = self.solver_class(data)

    def test_input_time_dimension_not_first(self):
        # time not the first dimension in the input
        solution = reference_solution('iris', 'equal')
        data = solution['sst']
        data.transpose((2, 1, 0))
        with pytest.raises(ValueError):
            solver = self.solver_class(data)

    def test_input_no_spatial_dimensions(self):
        # not enough dimensions in the input
        solution = reference_solution('iris', 'equal')
        data = solution['sst'][:, 0, 0]
        with pytest.raises(ValueError):
            solver = self.solver_class(data)

    def test_invalid_builtin_weights_value(self):
        # invalid weighting scheme name
        solution = reference_solution('iris', 'equal')
        data = solution['sst']
        with pytest.raises(ValueError):
            solver = self.solver_class(data, weights='invalid')

    def test_builtin_latitude_weights_with_missing_dimension(self):
        # latitude weights without latitude dimension
        solution = reference_solution('iris', 'latitude')
        data = solution['sst'][:, 0, :]
        data.remove_coord('latitude')
        with pytest.raises(ValueError):
            solver = self.solver_class(data, weights='coslat')

    def test_builtin_area_weights_with_missing_dimension(self):
        # latitude weights without latitude dimension
        solution = reference_solution('iris', 'area')
        data = solution['sst'][:, :, 0]
        data.remove_coord('longitude')
        with pytest.raises(ValueError):
            solver = self.solver_class(data, weights='area')

    def test_multiple_time_dimensions(self):
        # multiple dimensions representing time
        solution = reference_solution('iris', 'equal')
        data = solution['sst']
        lon = data.coord('longitude')
        lon.rename('time')
        lon.units = 'days since 1999-01-01 00:00:0.0'
        with pytest.raises(ValueError):
            solver = self.solver_class(data)


# ----------------------------------------------------------------------------
# Constructor tests for the xarray interface


class TestConstructorXarray(EofsTest):
    """Test the error handling in the xarray interface constructor."""

    @classmethod
    def setup_class(cls):
        try:
            cls.solver_class = solvers['xarray']
        except KeyError:
            pytest.skip('missing dependencies required to test the '
                        'xarray interface')

    def test_wrong_input_type(self):
        # input of the wrong type
        solution = reference_solution('standard', 'equal')
        data = solution['sst']
        with pytest.raises(TypeError):
            solver = self.solver_class(data)

    def test_input_without_time_dimension(self):
        # no time dimension in the input
        solution = reference_solution('xarray', 'equal')
        data = solution['sst'][0]
        with pytest.raises(ValueError):
            solver = self.solver_class(data)

    def test_input_time_dimension_not_first(self):
        # time not the first dimension in the input
        solution = reference_solution('xarray', 'equal')
        data = solution['sst']
        data = data.transpose('longitude', 'latitude', 'time')
        with pytest.raises(ValueError):
            solver = self.solver_class(data)

    def test_input_no_spatial_dimensions(self):
        # not enough dimensions in the input
        solution = reference_solution('xarray', 'equal')
        data = solution['sst'][:, 0, 0]
        with pytest.raises(ValueError):
            solver = self.solver_class(data)

    def test_invalid_builtin_weights_value(self):
        # invalid weighting scheme name
        solution = reference_solution('xarray', 'equal')
        data = solution['sst']
        with pytest.raises(ValueError):
            solver = self.solver_class(data, weights='invalid')

    def test_multiple_time_dimensions(self):
        # multiple dimensions representing time
        solution = reference_solution('xarray', 'equal')
        data = solution['sst']
        data.coords['longitude'].attrs['axis'] = 'T'
        with pytest.raises(ValueError):
            solver = self.solver_class(data)
